/*
 * spin.c  --  Model checker core routines
 *
 * NOTES:
 * 
 *  This file is part of the dSpin3.2 distribution.
 *  Original code written by Gerard Holzmann <gerard@research.bell-labs.com>.
 *  dSpin extension written by Radu Iosif <iosif@athena.polito.it>.
 */


#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <signal.h>


/* Hash function type */
typedef void (* hash_t)(unsigned char *, long);


#if defined(FULLSTACK) && defined(BITSTATE)
struct H_el * FreeList = (struct H_el *)0;
#endif


#ifdef COLLAPSE
static uchar scratch [POOLSZ];
#endif /* COLLAPSE */

/* Needed by linearize() */
static uchar lin [POOLSZ];

/* Needed by _delheap() and perm_heap() */
#define inside(i, k)				     \
(q_proc(i) == -1 && q_offset(i) >= h_offset(k) &&    \
 q_offset(i) + q_size(i) <= h_offset(k) + h_size(k))


void * emalloc(unsigned int n)
{
  register void *tmp;

  if (n == 0)
    return NULL;

  tmp = malloc(n);

  if (!tmp) {
    printf("pan: out of memory (emalloc)\n");
    printf("needed %d bytes more\n", n);
    wrapup(0);
  }

  memcnt += (double) n;
  memset((char *)tmp, 0, n);

  return tmp;
}


void * erealloc(void *p, unsigned int n, int inc)
{
  register void *tmp;

  if (n == 0)
    return NULL;

  tmp = realloc(p, n);

  if (!tmp) {
    printf("pan: out of memory (erealloc)\n");
    printf("needed %d bytes more\n", inc);
    wrapup(0);
  }

  if (n > inc && inc > 0)
    memset((char *)tmp + n - inc, 0, inc); 

  memcnt += (double) inc;
  return tmp;
}


void efree(void * tmp, int dec)
{
  if (!tmp || (memcnt < dec)) {
    printf("%s %d: cannot happen (efree)\n", __FILE__, __LINE__);
    exit(1);
  }

  memcnt -= (double) dec;
  free(tmp);
}


#ifdef POSYMM
int s_cmp(scalar_t *s1, scalar_t *s2)
{
  register int i;
  
  for (i = 0; (i < s1->len) && (i < s2->len); i ++)
    {
      if (s1->word[i] != s2->word[i])
	return (s1->word[i] < s2->word[i]) ? -1 : 1;
    }

  return (s1->len == s2->len) ? 0 : ((s1->len < s2->len) ? -1 : 1);
}


inline int ordered(scalar_t *s1, scalar_t *s2)
{
  return (s_cmp(s1, s2) <= 0);
}


void s_free(scalar_t *s)
{
  if (s->word) {
    efree(s->word, s->len * sizeof(uchar));
    s->word = (uchar *) 0;
  }

  s->len = 0;
}


void s_copy(scalar_t *d, scalar_t *s)
{
  if (d->len == s->len) {
    memcpy(d->word, s->word, d->len * sizeof(uchar));
    return;
  }

  s_free(d);
  
  d->len = s->len;
  d->word = (uchar *) emalloc(d->len * sizeof(uchar));
  memcpy(d->word, s->word, d->len * sizeof(uchar));
}


void s_push(scalar_t *s, uchar x)
{
  s->len ++;
  s->word = erealloc(s->word, s->len * sizeof(uchar), sizeof(uchar));
  s->word[s->len - 1] = x;
}
#endif /* POSYMM */


inline void q_copy(q_desc *d, q_desc *s)
{
  d->_offset = s->_offset;
  d->_size   = s->_size;
  d->_proc   = s->_proc;

#ifdef POSYMM
  s_copy(&d->_scalar, &s->_scalar);
#endif /* POSYMM */
}


inline void q_delete(int k)
{
  q_offset(k) = -1;
  q_size(k) = 0;
  q_proc(k) = 0;
  
#ifdef POSYMM
  s_free(&q_scalar(k));
#endif /* POSYMM */
}


inline void h_copy(h_desc *d, h_desc *s)
{
  d->_offset = s->_offset;
  d->_size   = s->_size;
  d->_range  = s->_range;
  d->_type   = s->_type;

#ifdef GC0
  d->_ref = s->_ref;
#endif /* GC0 */

#ifdef POSYMM
  s_copy(&d->_scalar, &s->_scalar);
#endif /* POSYMM */
}


inline void h_delete(int k)
{
  h_offset(k) = -1;
  h_size(k) = 0;
  h_range(k) = 0;
  h_type(k) = 0;

#ifdef GC0
  h_ref(k) = 0;
#endif /* GC0 */

#ifdef POSYMM
  s_free(&h_scalar(k));
#endif /* POSYMM */
}


inline void shift_queues(long threshold, int size)
{
  register int i, j;
  
  for (i = 0, j = 0; i < now._nr_qs; i ++, j ++)
    {
      for (; j < MAXQ && q_free(j); j ++);

      if (j == MAXQ) {
	printf("%s %d: cannot happen (shift_queues)\n", __FILE__, __LINE__);
	exit(1);
      }

      if (q_proc(j) == -1 && q_offset(j) > threshold)
	q_offset(j) += size;
    }
}


inline void shift_heap(long threshold, int size)
{
  register int i;

#ifdef POSYMM
  for (i = 0; i < MAXHEAP && !h_free(i); i ++) {
#else
  for (i = 0; i < MAXHEAP; i ++) {
#endif /* POSYMM */
    if (h_offset(i) > threshold)
      h_offset(i) += size;
  }
}


/* Stack frame management */

inline void push(int h, int sz)
{
#ifdef __ALIGN
  if (sz % WS)
    {
      printf("unaligned stack frame (push)\n");
      exit(1);
    }
#endif /* __ALIGN */

  if (this->_sp + sz > STACKSZ)
    {
      char str[256];	 
      
      sprintf(str, "pid %d: stack overflow, recompile with -DSTACKSZ>%d",
	      h, STACKSZ);
      Uerror(str);
    }

  pptr(h)->_sp += sz;
}


inline void pop(int h, int sz)
{
#ifdef __ALIGN
  if (sz % WS)
    {
      printf("unaligned stack frame (pop)\n");
      exit(1);
    }
#endif /* __ALIGN */

  if (pptr(h)->_sp < sz)
    {
      char str[256];

      sprintf(str, "pid %d: stack underflow", h);
      Uerror(str);
    }

  pptr(h)->_sp  -= sz;
}


/*
 * Pointer encoding and decoding routines. It works only for
 * 32-bit integer architectures so don't try this at home! :-)
 */

inline uint mkptr(uint off, uchar pid, uchar tag)
{
  switch (tag)
    {
    case LPTR:
      {     
	P *p;

	if (pid >= now._nr_pr)
	  fatal("invalid pid (lptr)");

	p = now.pool[pid];

	if (pid >= (1 << 6))
	  fatal("overlapping pid value (lptr)");

	if (off >= (1 << 24))
	  fatal("overlapping offset value (lptr)");
	
	return (tag << 30) | (pid << 24) | (p->_bp + off);
      }

    case GPTR:
    case HPTR:

      if (off >= (1 << 30))
	fatal("overlapping offset value (gptr)");

      return (tag << 30) | off;

    default:
      return (tag << 30);
    }
}


inline uchar *deptr(uint ptr, long inc)
{
  if (ptr == DPTR)
    Uerror("object was deleted");

  switch (ptr >> 30)
    {
    case LPTR:
      {
	int pid = (ptr & (63 << 24)) >> 24;
	int off = ptr & ((1 << 24) - 1);
	P *p;

	if (pid < 0 || pid >= now._nr_pr)
	  fatal("invalid pid (lptr)\n");

	p = now.pool[pid];

	if (off < 0 || off >= p->_sp)
	  fatal("invalid offset (lptr)\n");

	return p->stk + off;
      }

    case GPTR:
      {
	int off = ptr & ((1 << 30) - 1);

	if (off < 0 || off >= now._hsz)
	  fatal("invalid offset (gptr)\n");

	return now.heap + off;
      }

    case HPTR:
      {
	char str[256];
	int k = ptr & ((1 << 30) - 1);

	if (k < 0 || k >= MAXHEAP)
	  fatal("invalid index (hptr)\n");

	if (h_free(k)) {
	  sprintf(str, "object %d was deleted, cannot dereference", k); 
	  Uerror(str);
	}

	if (inc >= h_range(k)) {
	  sprintf(str, "reference to object %d out of range", k);
	  Uerror(str);
	}

	return now.heap + h_offset(k) + inc;
      }

    case NPTR:
      Uerror("attempt to dereference null pointer");

    default:
      fatal("bad pointer tag\n");
    }

  /* never reached */
  return (uchar *)0;
}


/*
 * Needed by delheap().
 */

inline int d_index(uint ptr)
{
  int k;

  if (ptr == DPTR)
    Uerror("object was deleted");

  switch (ptr >> 30)
    {
    case HPTR:
      break;

    case NPTR:
      Uerror("attempt to delete null pointer");
      
    default:
      Uerror("attempt to delete static variable");
    }

  k = ptr & ((1 << 30) - 1);

  if (k < 0 || k >= MAXHEAP)
    fatal("invalid heap index\n");

  if (h_free(k))
    Uerror("object was deleted");

  return k;
}


/* Function call management */

inline void call(int h, int ot, int tt)
{
  P *p;

  push(h, framesz);
  p = pptr(h);

  __ot(p) = p->_t;
  __tt(p) = p->_p;
  __bp(p) = p->_bp;

  p->_t   = (ot_t)ot;
  p->_p   = (tt_t)tt;
  p->_bp  = p->_sp;
}


inline void ret(int h, int v, int f)
{
  P *p = pptr(h);

  p->_sp = p->_bp;
  p->_t  = __ot(p);
  p->_p  = __tt(p);
  p->_bp = __bp(p);
  p->_rf = f;
  p->_rv = v;

  pop(h, framesz);
}


/*
 * Function pointers managing routines.
 * Assume we have a 32-bit unsigned integer.
 */

inline void cptr(int h, uint ptr)
{
  int ot = ptr >> 16;
  int tt = ptr & ((1 << 16) - 1);

  call(h, ot, tt);
}


inline uint fptr(int ot, int tt)
{
  if (ot >= (1 << 16))
    fatal("overlapping ot value (fptr)");

  if (tt >= (1 << 16))
    fatal("overlapping tt value (fptr)");

  return (ot << 16) | tt;
}


inline int fpot(uint ptr)
{
  return ptr >> 16;
}


/*
 * Function call using stack pollution
 */

inline void poll(int h, int ot, int tt)
{
  P *p = pptr(h);

  p->_t   = (ot_t)ot;
  p->_p   = (tt_t)tt;
  p->_sp  = p->_bp; 
}


inline void polp(int h, uint ptr)
{  
  int ot = ptr >> 16;
  int tt = ptr & ((1 << 16) - 1);

  poll(h, ot, tt);
}


inline int h_index(uint ptr)
{
  int k;
  
  if ((ptr == DPTR) || (HPTR != ptr >> 30))
    return -1;

  k = ptr & ((1 << 30) - 1);
  
  if (k < 0 || k >= MAXHEAP)
    fatal("invalid heap index (h_index)\n");

  return k;
}


/* Garbagge collection */
#ifndef NOGC
static long gc_runs = 0;


#ifdef GC0 /* reference counting GC */
inline void gc_inc(uint ptr)
{
  int k = h_index(ptr);

  if (k >= 0)
    h_ref(k) ++;
}


inline void gc_inc_back(uint ptr)
{
  int k = h_index(ptr);

  if (k >= 0)
    h_ref(k) --;
}


inline void __gc_dec(uint ptr, int save)
{
  int k = h_index(ptr);

  if (k >= 0 && h_ref(k))
    {
      if (!h_free(k) && h_ref(k) == 1) 
	{
	  if (save) {
	    _sv_save(&now);
	    (trpt + 1) -> sv ++;
	  }
	  
	  h_ref(k) = 0;
	  gc_rec(k);
	  gc_runs ++;
	}
      else 
	h_ref(k) --;

      (trpt + 1) -> gc ++;
    } 
}


inline void gc_dec(uint ptr, int save)
{
  register int k;

  __gc_dec(ptr, save);

#ifdef POSYMM
  for (k = 0; k < MAXHEAP && !h_free(k);)
    if (!h_ref(k))
      _delheap(k);
    else
      k ++;
#else
  for (k = 0; k < MAXHEAP; k ++)
    if (!h_free(k) && !h_ref(k))
      _delheap(k);
#endif /* POSYMM */
}


inline void gc_dec_back(uint ptr)
{
  if (trpt -> gc)
    {
      if (trpt -> sv) {
	sv_restor();
	trpt -> sv --;
      }
      else {
	int k = h_index(ptr);
	
	if (k >= 0) 
	  h_ref(k) ++;
      }

      trpt -> gc --;
    }
}
#endif /* GC0 */


#ifdef GC1 /* mark and sweep GC */
inline void gc_scan()
{
  register int h, k;

  for (k = 0; k < p_len[0]; k ++)
    gc_mark(gg(p_map[0][k], uint));

  for (h = 0; h < now._nr_pr; h ++)
    {
      P *p   = pptr(h);
      int sz = p->_sp - p->_bp;
      int bp = p->_bp;
      int ot = p->_t;

      if (p->_rf & 1)
	gc_mark(p->_rv);

      while (1) 
	{
	  for (k = 0; k < p_len[ot+1] && p_map[ot+1][k] < sz; k ++)
	    gc_mark(getll(p, p_map[ot+1][k], bp, uint));

	  if (bp) 
	    {
	      ot  = getot(p, bp);
	      sz  = bp - getbp(p, bp);
	      bp -= sz;
	    }
	  else break;
	}
    }

  gc_runs ++;
}

static double gc_coef = 0;

inline void gc_sweep(int save)
{
  register int k;
  double dels = 0, reach = 0;

  for (k = MAXHEAP - 1; k >= 0; k --)
    {
      if (!h_free(k)) 
	{ 
	  if (h_mark[k] == -1) 
	    {
	      if (save && !(trpt + 1) -> gc) 
		{
		  (trpt + 1) -> gc ++;
		  _sv_save(&now);
		}
	      
	      dels ++;
	      _delheap(k);
	    }
	  else h_mark[k] = -1;
	  reach ++;
	}
    }

  h_ordinal = 0;

  if (reach)
    gc_coef += dels / reach;
}
#endif /* GC1 */
#endif /* NOGC */


/* Symmetry reductions */
#ifdef POSYMM
void symm_get_scalar(scalar_t *d, int h)
{
  register int k;
  P *p = pptr(h);
  int sz = p->_sp - p->_bp;
  int bp = p->_bp;
  int ot = p->_t; /* proctype */
  int tt = p->_p; /* state */

  d->len = 0;
  d->word = (uchar *) 0;

  while (1)
    {
      s_push(d, ot);
      s_push(d, tt);

      if (bp)
	{
	  ot  = getot(p, bp);
	  tt  = gettt(p, bp);
	  sz  = bp - getbp(p, bp);
	  bp -= sz;
	}
      else break;
    }

  s_push(d, (uchar) h);
}
#endif /* POSYMM */


#ifdef FULLSYMM
void perm_ptr(int *perm, int len)
{
  register int j, k;

  for (k = 0; k < p_len[0]; k ++)
    {
      int i = h_index(gg(p_map[0][k], uint));
      if (0 <= i && i < len)
	gg(p_map[0][k], uint) = mp(perm[i], 0, HPTR);
    }

  for (k = 0; k < MAXHEAP && !h_free(k); k ++)
    fields_ptr_perm(k, perm, len);

  for (j = 0; j < now._nr_pr; j ++)
    {
      P *p   = now.pool[j];
      int sz = p->_sp - p->_bp;
      int bp = p->_bp;
      int ot = p->_t;
      int i;

      if (p->_rf & 1)
	{
	  i = h_index(p->_rv);
	  if (0 <= i && i < len)
	    p->_rv = mp(perm[i], 0, HPTR);
	}
      
      while (1)
	{
	  for (k = 0; k < p_len[ot+1] && p_map[ot+1][k] < sz; k ++)
	    {
	      i = h_index(getll(p, p_map[ot+1][k], bp, uint));
	      if (0 <= i && i < len)
		getll(p, p_map[ot+1][k], bp, uint) = mp(perm[i], 0, HPTR);
	    }

	  if (bp) 
	    {
	      ot  = getot(p, bp);
	      sz  = bp - getbp(p, bp);
	      bp -= sz;
	    }
	  else break;
	}
    }
}


void perm_chan(int *perm, int len)
{
  register int j, k;

  for (k = 0; k < c_len[0]; k ++)
    {
      int i = gg(c_map[0][k], uint) - 1;
      if (0 <= i && i < len)
	gg(c_map[0][k], uint) = perm[i] + 1;
    }

  for (k = 0; k < MAXHEAP && !h_free(k); k ++)
    fields_chan_perm(k, perm, len);

  for (j = 0; j < now._nr_pr; j ++)
    {
      P *p   = now.pool[j];
      int sz = p->_sp - p->_bp;
      int bp = p->_bp;
      int ot = p->_t;
      int i;

      if (p->_rf & 2)
	{
	  i = p->_rv - 1;
	  if (0 <= i && i < len)
	    p->_rv = perm[i] + 1;
	}

      while (1)
	{
	  for (k = 0; k < c_len[ot+1] && c_map[ot+1][k] < sz; k ++)
	    {
	      int i = getll(p, c_map[ot+1][k], bp, uchar) - 1;
	      if (0 <= i && i < len)
		getll(p, c_map[ot+1][k], bp, uchar) = perm[i] + 1;
	    }
	  
	  if (bp)
	    {
	      ot  = getot(p, bp);
	      sz  = bp - getbp(p, bp);
	      bp -= sz;
	    }
	  else break;	
	}      
    }
}


inline int copy_perm(int *perm, int reset)
{
  register int i, j;

  memset((char *)perm, 0, MAXHEAP * sizeof(int));

  for (i = 0, j = 0; i < MAXHEAP; i ++)
    if (h_mark[i] != -1) 
      {
	perm[j ++] = h_mark[i];

	if (reset)
	  h_mark[i] = -1;
      }

  if (reset)
    h_ordinal = 0;

  /* for (i = 0; i < j; i ++) printf("(%d->%d) ", i, perm[i]); printf("\n"); */

  return j;
}


inline void inverse_perm(int *perm, int *inv, int len)
{
  register int i;
  
  for (i = 0; i < len; i ++)
    inv[perm[i]] = i;
}


inline int iden_perm(int *perm, int len)
{
  register int i;
  
  for (i = 0; i < len; i ++)
    if (perm[i] != i)
      return 0;

  return 1;
}


int perm_qs(int *q, int l, int *u, int *owner)
{
  int x[MAXHEAP];
  int y[MAXHEAP];
  register int i, j, k;

  if (!now._nr_qs)
    return 0;

  memset((char *)x, 0, MAXHEAP * sizeof(int));
  memset((char *)y, 0, MAXHEAP * sizeof(int));

  for (i = 0, j = 0; i < l; i ++)
    if (inside(j, i))
      {
	x[i] = j ++;
	while (j < now._nr_qs && inside(j, i)) 
	  j ++;

	y[i] = j;
      }

  if (!j)
    return 0;

  for (i = 0, k = 0; i < l; i ++)
    for (j = x[q[i]]; j < y[q[i]]; j ++, k ++)
      {
	u[j] = k;
	owner[u[j]] = i;
      }

  /* for (i = 0; i < k; i ++) printf("[%d->%d] ", i, u[i]); printf("\n"); */

  return k;
}


void perm_heap(int *p, int l, int save)
{
  uchar tmp[HEAPSZ];
  h_desc ht[MAXHEAP];
  q_desc qt[MAXQ];
  int q[MAXHEAP];
  int mark[MAXHEAP];
  int u[MAXQ], v[MAXQ];
  int owner[MAXQ];
  int m = 0, n;
  register int i;

  if (!l || iden_perm(p, l))
    return;

  if (save && !(trpt + 1) -> gc)
    {
      (trpt + 1) -> gc ++;
      _sv_save(&now);
    }

  memset((char *)q, 0, MAXHEAP * sizeof(int));
  memset((char *)u, 0, MAXHEAP * sizeof(int));
  memset((char *)v, 0, MAXHEAP * sizeof(int));
  memset((char *)owner, 0, MAXHEAP * sizeof(int));
  memset((char *)mark, 0, MAXHEAP * sizeof(int));

  inverse_perm(p, q, l);

  for (i = 0, n = 0; i < l; i ++)
    {
      h_copy(&ht[i], &h_tab[q[i]]);
      mark[i] = h_mark[q[i]];

      ht[i]._offset = h_offset(0) + n;
      memcpy(tmp + ht[i]._offset, now.heap + h_offset(q[i]), h_size(q[i]));
      n += h_size(q[i]);
    }

  if (m = perm_qs(q, l, u, owner))
    {
      inverse_perm(u, v, m);

      for (i = 0; i < m; i ++)
	{
	  int k = owner[i];
	  q_copy(&qt[i], &q_tab[v[i]]);
	  qt[i]._offset = ht[k]._offset + q_offset(v[i]) - h_offset(q[k]);
	}
    }

  memcpy(now.heap + h_offset(0), tmp + h_offset(0), n);

  for (i = 0; i < l; i ++) 
    {
      h_copy(&h_tab[i], &ht[i]);
      h_mark[i] = mark[i];
    }

  for (i = 0; i < m; i ++) 
    q_copy(&q_tab[i], &qt[i]);

  perm_ptr(p, l);
  perm_chan(u, m);
}


inline void repnow(int save)
{
  int p[MAXHEAP], k, l;
  
  gc_scan();
  l = copy_perm(p, 1);
  perm_heap(p, l, save);
}
#endif /* FULLSYMM */


#ifndef NOSYMM
void symm_ptr_relocate(int threshold, int inc, int del)
#else
void remove_ptr_links(int del)
#endif /* NOSYMM */
{
  register int h, k;

  for (k = 0; k < p_len[0]; k ++) 
    {
      int i = h_index(gg(p_map[0][k], uint));
      
#ifndef NOSYMM
      if (i >= threshold)
	gg(p_map[0][k], uint) = mp(i + inc, 0, HPTR);
      else 
#endif /* NOSYMM */
	if (del >= 0 && i == del)
	  gg(p_map[0][k], uint) = DPTR;
    }

#ifndef NOSYMM
  for (k = 0; k < MAXHEAP && !h_free(k); k ++)
    fields_ptr_reloc(k, threshold, inc, del);
#else
  for (k = 0; k < MAXHEAP; k ++)
    if (!h_free(k)) fields_ptr_remove(k, del);
#endif /* NOSYMM */

  for (h = 0; h < now._nr_pr; h ++)
    {
      P *p   = pptr(h);
      int sz = p->_sp - p->_bp;
      int bp = p->_bp;
      int ot = p->_t;
      int i;

      if (p->_rf & 1)
	{
	  i = h_index(p->_rv);

#ifndef NOSYMM
	  if (i >= threshold)
	    p->_rv = mp(i + inc, 0, HPTR);
	  else
#endif /* NOSYMM */
	    if (del >= 0 && i == del)
	      p->_rv = DPTR;
	}

      while (1)
	{
	  for (k = 0; k < p_len[ot+1] && p_map[ot+1][k] < sz; k ++)
	    {
	      i = h_index(getll(p, p_map[ot+1][k], bp, uint));

#ifndef NOSYMM
	      if (i >= threshold)
		getll(p, p_map[ot+1][k], bp, uint) = mp(i + inc, 0, HPTR);
	      else
#endif /* NOSYMM */
		if (del >= 0 && i == del)
		  getll(p, p_map[ot+1][k], bp, uint) = DPTR;
	    }
	  
	  if (bp) 
	    {
	      ot  = getot(p, bp);
	      sz  = bp - getbp(p, bp);
	      bp -= sz;
	    }
	  else break;	  
	}
    }
}


#ifndef NOSYMM
void symm_chan_relocate(int threshold, int inc, int del)
#else
void remove_chan_links(int del)
#endif /* NOSYMM */
{
  register int h, k;

  for (k = 0; k < c_len[0]; k ++) 
    {
#ifndef NOSYMM
      if (gg(c_map[0][k], uchar) >= threshold)
	gg(c_map[0][k], uchar) += inc;
      else 
#endif /* NOSYMM */
	if (del && gg(c_map[0][k], uchar) == del)
	  gg(c_map[0][k], uint) = 0;
    }

#ifndef NOSYMM
  for (k = 0; k < MAXHEAP && !h_free(k); k ++)
    fields_chan_reloc(k, threshold, inc, del);
#else
  for (k = 0; k < MAXHEAP; k ++)
    if (!h_free(k)) fields_chan_remove(k, del);
#endif /* NOSYMM */

  for (h = 0; h < now._nr_pr; h ++)
    {
      P *p   = pptr(h);
      int sz = p->_sp - p->_bp;
      int bp = p->_bp;
      int ot = p->_t;

      if (p->_rf & 2)
	{
#ifndef NOSYMM
	  if (p->_rv >= threshold)
	    p->_rv += inc;
	  else
#endif /* NOSYMM */
	    if (del && p->_rv == del)
	      p->_rv = 0;
	}

      while (1)
	{
	  for (k = 0; k < c_len[ot+1] && c_map[ot+1][k] < sz; k ++)
	    {	  
#ifndef NOSYMM
	      if (getll(p, c_map[ot+1][k], bp, uchar) >= threshold)
		getll(p, c_map[ot+1][k], bp, uchar) += inc;
	      else 
#endif /* NOSYMM */
		if (del && getll(p, c_map[ot+1][k], bp, uchar) == del)
		  getll(p, c_map[ot+1][k], bp, uchar) = 0;
	    }
	  
	  if (bp)
	    {
	      ot  = getot(p, bp);
	      sz  = bp - getbp(p, bp);
	      bp -= sz;
	    }
	  else break;	  
	}
    }
}


/*
 * Keeps the memcmp()'s convention for return value
 */

int _statecmp(State *s1, State *s2)
{
  register int i;
  register int t;

  if ((t = s1->_hsz - s2->_hsz))
    return t;

  if ((t = s1->_nr_pr - s2->_nr_pr))
    return t;

  if ((t = memcmp((char *)s1->heap, (char *)s2->heap, s1->_hsz)))
    return t;
    
  for (i = 0; i < s1->_nr_pr; i ++) 
    {
      if ((t = s1->pool[i]->_sp - s2->pool[i]->_sp))
	return t;

      if ((t = memcmp((char *)s1->pool[i], 
		      (char *)s2->pool[i], 
		      p_sz(s1->pool[i]))))
	return t;
    }

  return 0;
}


long _statesz(State *s)
{
  register int i;
  register long sz = 0;

  sz += 3 * sizeof(uchar);

#ifndef NOFAIR
  sz += NFAIR * sizeof(uchar);
#endif

#ifdef HAS_LAST
  sz += sizeof(uchar);
#endif

#ifdef EVENT_TRACE
#if nstates_event < 256
  sz += sizeof(uchar);
#else
  sz += sizeof(unsigned short);
#endif
#endif

#if HEAPSZ < 65536
  sz += sizeof(unsigned short);
#else
  sz += sizeof(unsigned long);
#endif

  sz += s->_hsz;

  for (i = 0; i < s->_nr_pr; i ++)
    sz += p_sz(s->pool[i]);

  return sz;
}


/*
 * For debugging
 */

void _statedump(State *s)
{
  register int i, k;
  P * p;
  Q * q;

  printf("[");

  for (i = 0; i < s->_nr_pr; i ++) {
    p = s->pool[i];
    printf("(%d %d %d)", p->_pid, p->_t, p->_p);
 }
  /*
  for (i = 0; i < s->_nr_qs; i ++) {
    q = qptr(i);
    printf("{%d %d}", i, q->Qlen);
  }
  */
  printf("]");
}


void _procdup(Stack *s, int h)
{
  P *p	 = pptr(h);
  int sz = p_sz(p);

  S_P(s) = (P *) emalloc(sz);
  memcpy((char *)S_P(s), (char *)p, sz);
}


void _queuedup(Stack *s, int k)
{
  int sz = q_size(k);

  S_Q(s) = (Q *) emalloc(sz);
  memcpy((char *)S_Q(s), (char *)qptr(k), sz);
}


State * _statedup(State *s)
{
  register int i;
  State * t;
  
  t = (State *) emalloc(sizeof(State));
  
  t->_nr_pr = s->_nr_pr;
  t->_nr_qs = s->_nr_qs;
  t->_a_t   = s->_a_t;
  t->_hsz   = s->_hsz;

  t->heap = (uchar *) emalloc(s->_hsz);
  memcpy((char *)t->heap, (char *)s->heap, s->_hsz);

  t->pool = (P **) emalloc(s->_nr_pr * sizeof(P *));

  for (i = 0; i < s->_nr_pr; i ++) 
    {
      t->pool[i] = (P *) emalloc(p_sz(s->pool[i]));
      memcpy((char *)t->pool[i], (char *)s->pool[i], p_sz(s->pool[i]));
    }

  return t;
}


void _statefree(State *s)
{
  register int i;

  if (s->_hsz)
    efree(s->heap, s->_hsz);

  for (i = 0; i < s->_nr_pr; i ++)
    efree(s->pool[i], p_sz(s->pool[i]));

  if (s->_nr_pr)
    efree(s->pool, s->_nr_pr * sizeof(P *));

  efree(s, sizeof(State));
}


void _statecpy(State *d, State *s)
{
  register int i;

  d->_nr_pr = s->_nr_pr;
  d->_nr_qs = s->_nr_qs;
  d->_a_t   = s->_a_t;
  d->_hsz   = s->_hsz;

  memcpy((char *)d->heap, (char *)s->heap, s->_hsz);

  for (i = 0; i < s->_nr_pr; i ++)
    memcpy((char *)d->pool[i], (char *)s->pool[i], p_sz(s->pool[i]));
}


int _addproc(int n)
{
  int h = now._nr_pr;

#ifndef INLINE
  if (TstOnly)
    return (h < MAXPROC);
#endif

  if (h >= MAXPROC)
    Uerror("too many processes");  

  if (fairness && ((int) now._nr_pr + 2 >= (8 * NFAIR) / 2))
    {	
      printf("Error: too many processes -- current");
      printf(" max is %d procs (-DNFAIR=%d)\n", (8*NFAIR)/2 - 2, NFAIR);
      printf("\trecompile with -DNFAIR=%d\n", NFAIR+1);
      exit(1);
    }

  now._nr_pr ++;

  pptr(h) = (P *) emalloc(sizeof(P));
  pptr(h)->_type = n;
  pptr(h)->_t = n;

  if (BASE > 0 && h > 0)
    pptr(h)->_pid = h - BASE;
  else
    pptr(h)->_pid = h;

  hmax = max(hmax, _statesz(&now));
  
  return h;
}


#ifdef POSYMM
int _addq(int t, int h, uchar n, int size, int align, scalar_t *s)
#else
int _addq(int t, int h, uchar n, int size, int align)
#endif /* POSYMM */
{
  int i, k, l;
  Q *q;

  for (k = 0; k < MAXQ && !q_free(k); k ++);

  if (k == MAXQ)
    Uerror("too many queues");

  now._nr_qs ++;

  switch (t)
    {
    case Q_GLOBAL:

#ifdef __ALIGN
      if (align)
	now._hsz += PAD(now._hsz);
#endif /* __ALIGN */

      q_offset(k) = now._hsz;
      q_size(k) = size;
      q_proc(k) = -1;
      now._hsz += size;

      if (now._hsz >= HEAPSZ) 
	{
	  char str[256]; 
	  sprintf(str,"HEAPSZ too small, recompile with -DHEAPSZ>%d",HEAPSZ);
	  Uerror(str);
	}

      break;
      
    case Q_HEAP:

#ifdef POSYMM
      for (l = 0; l < k && ordered(s, &q_scalar(l)); l ++);
      
      if (l < k) 
	{
	  symm_chan_relocate(l + 1, 1, -1);
	  
	  for (i = k - 1; i >= l; i --)
	    q_copy(&q_tab[i + 1], &q_tab[i]);
      
	  k = l;
	}
      
      s_copy(&q_scalar(k), s);
#endif /* NOSYMM */

      /*
       * Heap allocated queues should start with aligned adresses 
       * because object's offsets and sizes are aligned by _addheap
       */

#ifdef __ALIGN
      size += PAD(size);
#endif /* __ALIGN */

      q_offset(k) = h_offset(h) + h_size(h);
      q_size(k) = size;
      h_size(h) += size;
      now._hsz += size;

      if (now._hsz >= HEAPSZ) 
	{
	  char str[256]; 
	  sprintf(str,"HEAPSZ too small, recompile with -DHEAPSZ>%d",HEAPSZ);
	  Uerror(str);
	}

      shift_queues(q_offset(k), size);
      shift_heap(h_offset(h), size);

      memmove(now.heap + h_offset(h) + h_size(h),
	      now.heap + h_offset(h) + h_size(h) - size,
	      now._hsz - h_offset(h) - h_size(h) + size);

      q_proc(k) = -1;
      break;

    case Q_LOCAL:
      /*
       * Don't need to align local queues because we know 
       * that _sp will allways point to an aligned address.
       */

      q_offset(k) = pptr(h)->_sp;
      q_size(k) = size;

#ifdef __ALIGN
      size += PAD(size);
#endif /* __ALIGN */

      push(h, size);
      q_proc(k) = h;
      break;

    default:
      fatal("bad queue tag");
    }
  
  q = qptr(k);
  
  memset((char *) q, 0, size);
  q->_t = n;
  
  hmax = max(hmax, _statesz(&now));

  return k + 1;
}


void _sv_save(State *won)
{
  int i;

  if (!svtack->nxt)
    {  
      svtack->nxt = (Svtack *) emalloc(sizeof(Svtack));
      svtack->nxt->lst = svtack;
      svmax++;
    } 
	
  svtack = svtack->nxt;
  svtack->body = _statedup(won);

  for (i = 0; i < MAXQ; i ++) 
    {
      if (!svtack->q_tab[i])
	svtack->q_tab[i] = (q_desc *) emalloc(sizeof(q_desc));
      
      q_copy(svtack->q_tab[i], &q_tab[i]);
    }

  for (i = 0; i < MAXHEAP; i ++) 
    {
      if (!svtack->h_tab[i])
	svtack->h_tab[i] = (h_desc *) emalloc(sizeof(h_desc));
      
      h_copy(svtack->h_tab[i], &h_tab[i]);
    }

#if SYNC
  svtack->o_boq = boq;
#endif

#ifdef DEBUG
  fprintf(stdout, "%d: _sv_save ", depth);
  _statedump(won);
  fprintf(stdout, "\n");
#endif
}


void _sv_restor()
{
  int i;

  if (!svtack || !svtack->lst) {
    printf("%s %d: cannot happen (_sv_restor)\n", __FILE__, __LINE__);
    exit(1);
  }

#if SYNC
  boq = svtack->o_boq;
#endif

  for (i = 0; i < MAXQ; i ++)
    q_copy(&q_tab[i], svtack->q_tab[i]);

  for (i = 0; i < MAXHEAP; i ++)
    h_copy(&h_tab[i], svtack->h_tab[i]);
 
  _statecpy(&now, svtack->body);
  _statefree(svtack->body);

#ifdef DEBUG
  fprintf(stdout , "%d: _sv_restor ", depth);
  _statedump(&now);
  fprintf(stdout , "\n");
#endif

  svtack = svtack->lst;
}


void _p_restor(int h)
{
  int i, j;
  uchar qs[MAXQ];

  if (!stack || !stack->lst) {
    printf("%s %d: cannot happen (_p_restor)\n", __FILE__, __LINE__);
    exit(1);
  }

#ifndef XUSAFE
  p_name[h] = stack->o_name;
#endif

  now._nr_pr ++;  

  memcpy((char *)pptr(h), (char *)S_P(stack), stack->o_size);
  efree(S_P(stack), stack->o_size);

  this = pptr(h);

  if (BASE > 0 && h > 0)
    this->_pid = h-BASE;
  else
    this->_pid = h;

  i = stack->o_delqs;

  for (j = 0; j < i; j ++)
    qs[j] = stack->o_qs[j];

  stack = stack->lst;

  while (-- i >= 0)
    _q_restor(qs[i]);
}


void _q_restor(int k)
{
  if (!stack || !stack->lst) {
    printf("%s %d: cannot happen (_q_restor)\n", __FILE__, __LINE__);
    exit(1);
  }

#ifndef XUSAFE
  q_name[k]   = stack->o_name;
#endif

  q_proc(k)   = stack->o_proc;
  q_offset(k) = stack->o_offset;
  q_size(k)   = stack->o_size;

  now._nr_qs ++;

  memcpy((char *)qptr(k), (char *)S_Q(stack), stack->o_size);
  efree(S_Q(stack), stack->o_size);

  stack = stack->lst;
}


void _p_save(int h, int i, uchar *qs)
{
  int j;

  if (!stack->nxt) {
    stack->nxt = (Stack *) emalloc(sizeof(Stack));
    stack->nxt->lst = stack;
    smax ++;
  }

  stack = stack->nxt;

  _procdup(stack, h);
  
  stack->o_size	 = p_sz(pptr(h)); 
  stack->o_delqs = i;
  
  for (j = 0; j < i; j ++)
    stack->o_qs[j] = qs[j];

#ifndef XUSAFE
  stack->o_name = p_name[h];
#endif
}


void _q_save(int k)
{
  if (!stack->nxt) {
    stack->nxt = (Stack *) emalloc(sizeof(Stack));
    stack->nxt->lst = stack;
    smax ++;
  }

  stack = stack->nxt;

  _queuedup(stack, k);

  stack->o_proc	  = q_proc(k);
  stack->o_offset = q_offset(k);
  stack->o_size	  = q_size(k);

#ifndef XUSAFE
  stack->o_name = q_name[k];
#endif
}


void _p_q_save(int h, int k)
{
  if (!stack->nxt)
    {
      stack->nxt = (Stack *) emalloc(sizeof(Stack));
      stack->nxt->lst = stack;
      smax ++;
    }

  stack = stack->nxt;

  _procdup(stack, h);

  stack->o_size	 = p_sz(pptr(h)); 

  if (!stack->nxt)
    {
      stack->nxt = (Stack *) emalloc(sizeof(Stack));
      stack->nxt->lst = stack;
      smax ++;
    }

  stack = stack->nxt;

  _queuedup(stack, k);

  stack->o_proc	  = q_proc(k);
  stack->o_offset = q_offset(k);
  stack->o_size	  = q_size(k);

#if SYNC
  svtack->o_boq = boq;
#endif
}


void _p_q_restor(int h, int k)
{
  if (!stack || !stack->lst || !stack->lst->lst) {
    printf("%s %d: cannot happen (_p_q_restor)\n", __FILE__, __LINE__);
    exit(1);
  }

#if SYNC
  boq = svtack->o_boq;
#endif

  q_offset(k) = stack->o_offset;
  q_size(k)   = stack->o_size;
  q_proc(k)   = stack->o_proc;

  memcpy((char *)qptr(k), (char *)S_Q(stack), stack->o_size);
  efree(S_Q(stack), stack->o_size);

  stack = stack->lst;

  memcpy((char *)pptr(h), (char *)S_P(stack), stack->o_size);
  efree(S_P(stack), stack->o_size);

  stack = stack->lst;
}


void _delproc(int save, int h)
{
  int i = 0, j, k;
  uchar qs[MAXQ];

  for (j = 0, k = 0; j < now._nr_qs; j ++, k ++)
    {
      for (; k < MAXQ && q_free(k); k ++);

      if (k == MAXQ) {
	printf("%s %d: cannot happen (_delproc)\n", __FILE__, __LINE__);
	exit(1);
      }

      if (q_proc(k) == h)
	{
	  _delq(Q_LOCAL, k, save);
	  qs[i ++] = k;
	}
    }

  if (save)
    _p_save(h, i, qs);

  now._nr_pr --;
}


void _delq(int t, int k, int save)
{
  int i;

  if (k < 0 || q_free(k)) {
    printf("%s %d: cannot happen (_delq)\n", __FILE__, __LINE__);
    exit(1);
  }

  now._nr_qs --;

  if (save)
    _q_save(k);

  memset(qptr(k), 0, q_size(k));

  if (t == Q_HEAP) {
#ifndef NOSYMM
    symm_chan_relocate(k + 2, -1, k + 1);

    for (i = k; i < MAXQ - 1 && !q_free(i); i ++)
      q_copy(&q_tab[i], &q_tab[i + 1]);

    k = i;
#else
    remove_chan_links(k + 1);
#endif /* NOSYMM */
  }

  q_delete(k);
}


#ifdef POSYMM
int _addheap(int size, uint type, scalar_t *s)
#else
int _addheap(int size, uint type)
#endif /* POSYMM */
{
  int newsz = size + sizeof(uint);
  int i, k, l;

  for (k = 0; k < MAXHEAP && !h_free(k); k ++);

  if (k == MAXHEAP)
    Uerror("too many objects");

#ifdef __ALIGN
  now._hsz += PAD(now._hsz);
  newsz += PAD(newsz);
#endif /* __ALIGN */

  if (now._hsz + newsz >= HEAPSZ) 
    {
      char str[256];
      sprintf(str, "HEAPSZ too small, recompile with -DHEAPSZ>%d", HEAPSZ);
      Uerror(str);
    }

#ifdef POSYMM
  for (l = 0; l < k && ordered(s, &h_scalar(l)); l ++);
  if (l == k)
    {
      h_offset(l) = now._hsz;
    }
  else
    {
      shift_queues(h_offset(l), newsz);
      symm_ptr_relocate(l, 1, -1);

      for (i = k - 1; i >= l; i --)
	{
	  h_copy(&h_tab[i + 1], &h_tab[i]);
	  h_offset(i + 1) += newsz;
	}

      memmove(now.heap + h_offset(l) + newsz,
	      now.heap + h_offset(l),
	      now._hsz - h_offset(l));
    }

  s_copy(&h_scalar(l), s);
  k = l;
#else
  h_offset(k) = now._hsz;
#endif /* POSYMM */

  now._hsz += newsz;

  h_size(k) = newsz;
  h_range(k) = size;
  h_type(k) = type;

#ifdef GC0
  h_ref(k) = 1;
#endif /* GC0 */

  memset(now.heap + h_offset(k), 0, newsz);
  memcpy(now.heap + h_offset(k) + size, (char *) &type, sizeof(uint));

  hmax = max(hmax, _statesz(&now));

  return k;
}


void _delheap(int k)
{
  register int i, j;
  int freesz = h_size(k);

  for (i = 0, j = 0; i < now._nr_qs; i ++, j ++)
    {
      for (; j < MAXQ && q_free(j); j ++);

      if (j == MAXQ) {
	printf("%s %d: cannot happen (_delheap)\n", __FILE__, __LINE__);
	exit(1);
      }
      
      if (inside(j, k))
	_delq(Q_HEAP, j, 0);
    }

  shift_queues(h_offset(k), - freesz);

#ifndef NOSYMM
  symm_ptr_relocate(k + 1, -1, k);
#else
  remove_ptr_links(k);
  shift_heap(h_offset(k), - freesz);
#endif /* NOSYMM */

  memmove(now.heap + h_offset(k), 
	  now.heap + h_offset(k) + freesz, 
	  now._hsz - h_offset(k) - freesz);

  now._hsz -= freesz;
  
#ifndef NOSYMM
  for (i = k; i < MAXHEAP - 1 && !h_free(i + 1); i ++)
    {
      h_copy(&h_tab[i], &h_tab[i + 1]);
      h_offset(i) -= freesz;
    }

  k = i;
#endif /* NOSYMM */
  
  h_delete(k);
}


void assert(int a, char *s, int ii, int tt, Trans *t)
{
  if (!a && !noasserts)
    {	
      char bad[250];

      sprintf(bad, "assertion violated %s", s);      
      depth++; 
      trpt++;

      if (t) {
	trpt->pr = ii;
	trpt->st = tt;
	trpt->o_t = t;
      } 
      else {
	trpt->pr = (trpt-1)->pr;
	trpt->st = (trpt-1)->st;
	trpt->o_t = (trpt-1)->o_t;
      }
	
      uerror(bad);
      depth--; 
      trpt--;
    }
}


int qs_empty(void)
{	
  int i, j;

  for (i = 0, j = 0; i < now._nr_qs; i ++, j ++)
    {
      for (; j < MAXQ && q_free(j); j ++);

      if (j == MAXQ) {
	printf("%s %d: cannot happen (qs_empty)\n", __FILE__, __LINE__);
	exit(1);
      }

      if (q_size(j) > 0)
	return 0;
    }

  return 1;
}


#ifndef NOBOUNDCHECK
int Boundcheck(int x, int y, int a1, int a2, Trans *a3)
{
  assert((x >= 0 && x < y), "- invalid array index", a1, a2, a3);
  return x;
}
#endif /* NOBOUNDCHECK */


void _hinit()
{
#if defined(BITSTATE)
  SS = (uchar *) emalloc (1 << (ssize - 3));
#else 
#ifdef MA
#ifdef R_XPT
  r_xpoint();
#else
  dfa_init(MA + a_cycles);
#endif /* R_XPT */
#endif /* MA */

#if !defined(MA) || defined(COLLAPSE)
  H_tab = (struct H_el **) emalloc((1 << ssize) * sizeof(struct H_el *));
#endif
#endif /* BITSTATE */

  mask = ((1 << ssize) - 1);
}


inline int inc_code(uchar *cp, long om, int m, long z)
{
  register long *q;
  register long h;

  /* Attempt to code an empty vector */
  if (!om)
    return m;

  h = (om + sizeof(long) - 1) / sizeof(long);
  q = (long *)cp;

  do {
    if (m < 0) {
      m += m;
      m ^= z;
    }
    else 
      m += m;

    m ^= *q ++;
  } while (-- h > 0);

  return m;
}


inline int dec_code(uchar *cp, long om, int m, long z)
{
  register long *q;
  register long h;

  /* Attempt to code an empty vector */
  if (!om)
    return m;

  h = (om + sizeof(long) - 1) / sizeof(long);
  q = (long *)cp + h;

  do {
    if (m < 0) {
      m += m;
      m ^= z;
    }
    else
      m += m;

    m ^= * --q;
  } while (-- h > 0);

  return m;
}


/*
 * Normal hashing function 
 */

void s_hash(uchar * s, long n)
{
  register long z = (long) HASH_CONST[HASH_NR];
  register int m;

  m = inc_code(s, n, -1, z);

  j1 = (m ^ (m >> (8 * sizeof(long) - ssize))) & mask;
}


/*
 * Bitstate hashing, needed by _onstack_(zap|put|now)()
 */

void b_hash(uchar * s, long n)
{
  register long z = (long) HASH_CONST[HASH_NR];
  register int m;

  m = inc_code(s, n, -1, z);

  j1 = (m ^ (m >> (8 * sizeof(long) - (ssize-3)))) & ((1 << (ssize-3)) - 1);
}


/*
 * Bitstate double hashing function
 */

void d_hash(uchar * s, long n)
{
  register long z = (long) HASH_CONST[HASH_NR];
  register int m1, m2;

  m1 = inc_code(s, n, -1, z);
  m2 = dec_code(s, n, -1, z);

  J1 = (m1 ^ (m1 >> (8 * sizeof(long) - ssize))) & mask;
  J2 = (m2 ^ (m2 >> (8 * sizeof(long) - ssize))) & mask;

  if (!single) {
    j1 = J1 >> 3; j3 = (1 << (J1 & 7));
    j2 = J2 >> 3; j4 = (1 << (J2 & 7));	      
  }
  else {
    J1 = J1 ^ J2;
    j1 = 0; j3 = (1 << (J1 & 7));
    j2 = J1 >> 3; j4 = 1;
  }
}


#if defined(FULLSTACK) && defined(BITSTATE)
static inline struct H_el * grab_H_el(uchar * s, long n)
{
  register struct H_el *h, *last = 0;

  for (h = FreeList; h && h->tagged >= n; h = h->nxt)
    {	
      if (h->tagged == n)
	{	
	  if (last)
	    last->nxt = h->nxt;
	  else
	    FreeList = h->nxt;

	  h->tagged = 0;
	  h->nxt = 0;
#ifdef COLLAPSE
	  h->ln = 0;
#endif
	  memcpy((char *)&(h->state), s, n);

	  return h;
	}
      
      Fh ++; 
      last = h;
    }

  ngrabs ++;

  h = (struct H_el *) emalloc(sizeof(struct H_el) - sizeof(unsigned) + n);
  memcpy((char *)&(h->state), s, n);

  return h;
}

#else /* FULLSTACK && BITSTATE */


static inline struct H_el * grab_H_el(uchar * s, long n)
{
  register struct H_el * h;

  h = (struct H_el *) emalloc(sizeof(struct H_el) - sizeof(unsigned) + n);
  memcpy((char *)&(h->state), s, n);

  return h;
}
#endif /* defined(FULLSTACK) && defined(BITSTATE) */


static inline int bytes_no(unsigned long n, uchar * v)
{
  register int i = 0;

  if (n >= (1 << 8)) {
    i ++;
    *v ++ = (n >> 8) & 255;
  }

  if (n >= (1 << 16)) {
    i ++;
    *v ++ = (n >> 16) & 255;
  }

  if (n >= (1 << 24)) {
    i ++;
    *v ++ = (n >> 24) & 255;
  }

  return i;
}


static inline long lin_p(P * p, char * v)
{
  register long n = 0;
  unsigned int prlsz = offsetof(P, pad) - 1;

  /* Copy the prologue */
#ifdef NOCOMP
  v [n ++] = p->_pid;
#endif

  /* Skip pad, _sp, _bp and _rv */
  memcpy(v + n, (char *)p + 1, prlsz);
  n += prlsz;

  /* Copy the stack */
  memcpy(v + n, p->stk, p->_sp);
  n += p->_sp;
  
  return n;
}


#ifndef COLLAPSE

/*
 * Copy our state structure into a byte vector. 
 * State compression is performed during linearization.
 */

static long linearize(State * s)
{
  register long i = 0;
  register int h, d;
  P * p;

  /* Copy the prologue */
#ifdef NOCOMP

  memcpy(lin, (char *)&(s->_nr_pr), sizeof(uchar));
  memcpy(lin + i, (char *)&(s->_nr_qs), sizeof(uchar));
  memcpy(lin + i, (char *)&(s->_a_t), sizeof(uchar));
  i += 3 * sizeof(uchar);

#ifndef NOFAIR
  memcpy(lin + i, (char *)s->_cnt, NFAIR * sizeof(uchar));
  i += NFAIR * sizeof(uchar);
#endif /* NOFAIR */

#endif /* NOCOMP */

#ifdef HAS_LAST
  memcpy(lin + i, (char *)&(s->_last), sizeof(uchar));
  i += sizeof(uchar);
#endif /* HAS_LAST */

#ifdef EVENT_TRACE
#if nstates_event < 256
  memcpy(lin + i, (char *)&(s->_event), sizeof(uchar));
  i += sizeof(uchar);
#else
  memcpy(lin + i, (char *)&(s->_event), sizeof(unsigned short));
  i += sizeof(unsigned short);
#endif
#endif /* EVENT_TRACE */

  /* Copy the heap area */
  memcpy(lin + i, s->heap, s->_hsz);
  i += s->_hsz;
  
#ifdef NOCOMP
  memset(lin + i, 0, d = PAD(i));
  i += d;
#endif

  /* Copy the process pool */
  for (h = 0; h < s->_nr_pr; h ++) 
    {
      p = s->pool[h];
      i += lin_p(p, lin + i);
      
#ifdef NOCOMP
      memset(lin + i, 0, d = PAD(i));
      i += d;
#endif /* NOCOMP */
    }

#ifdef NOCOMP
  memset(lin + i, 0, d = PAD(i));
  i += d;
#endif

  lin [i ++] = i & 255;
  i += bytes_no(i, lin + i);

  /* Clear trailing garbage */
  for (h = 0; h < WS - 1; h ++)
    lin [i + h] = 0;

  return i;
}

#else /* COLLAPSE */

/*
 * Store state components into H_tab[].
 */

static unsigned long ordinal(char *v, long n, short tp)
{
  struct H_el *tmp, *ntmp;
  struct H_el *olst = (struct H_el *)0;
  long m;

  s_hash((uchar *)v, n);

  tmp = H_tab[j1];

  if (!tmp) {
    tmp = grab_H_el(v, n);
    H_tab[j1] = tmp;
  }
  else
    for (;; olst = tmp, tmp = tmp->nxt) 
      {
	m = memcmp((char *)&tmp->state, v, n);

	if (n == tmp->ln)
	  {
	    if (m == 0)
	      goto Done;
	    
	    if (m < 0)
	      {
	      Insert:

		ntmp = grab_H_el(v, n);
		ntmp->nxt = tmp;

		if (!olst)
		  H_tab[j1] = ntmp;
		else
		  olst->nxt = ntmp;

		tmp = ntmp;
		break;
	      }
	    else
	      if (!tmp->nxt)
		{
		Append:

		  tmp->nxt = grab_H_el(v, n);
		  tmp = tmp->nxt;
		  break;
		}
	    continue;
	  }

	if (n < tmp->ln)
	  goto Insert;
	else
	  if (!tmp->nxt)
	    goto Append;
      }

  m = ++ ncomps [tp];

#ifdef FULLSTACK
  tmp->tagged = m;
#else
  tmp->st_id  = m;
#endif

  tmp->ln = n;

 Done:
#ifdef FULLSTACK
  return tmp->tagged;
#else
  return tmp->st_id;
#endif
}


/*
 * Collapse process.
 */

static long col_p(int i, uchar * v)
{
  P *p = pptr(i);
  uchar *x;
  long j;
  int k;

  if (v)
    x = v;
  else
    x = scratch;

  j = lin_p(p, x);

  if (v)
    return j;
  
  for (k = 0; k < WS - 1; k ++)
    x [j + k] = 0;

  return ordinal(x, j, 2 + p->_type);
}


/*
 * Collapse queue.
 */

static long col_q(int i, char * v)
{
  Q *q = qptr(i);
  long j = q_size(i);
  uchar *x;
  int k;

  if (v)
    x = v;
  else
    x = scratch;

  j -= (q_max [q->_t] - q->Qlen) * (j - 2) / q_max [q->_t];
  memcpy(x, (char *)q, j);

  if (v)
    return j;

  for (k = 0; k < WS - 1;  k ++)
    x [j + k] = 0;

  return ordinal(x, j, 1);
}


/*
 * Collapse our state structure into a byte vector.
 */

static long linearize(State * s)
{
  register long i = 0, j = 0;
  register int p, q, k, l;
  unsigned long n;

  static unsigned char nbytes [513];
  static unsigned short nbytelen;

#ifndef SAFETY
  if (a_cycles) {
    memcpy(lin, (char *)&(s->_a_t), sizeof(uchar));
    i += sizeof(uchar);
  }

#ifndef NOFAIR
  if (fairness) {
    memcpy(lin + i, (char *)s->_cnt, NFAIR * sizeof(uchar));
    i += NFAIR * sizeof(uchar);
  }
#endif /* NOFAIR */
#endif /* SAFETY */

  nbytelen = 0;

#ifndef JOINPROCS
  for (k = 0; k < s->_nr_pr; k ++)
    {
      n = col_p(k, NULL);
      lin [i ++] = n & 255;
      i += nbytes [nbytelen] = bytes_no(n, lin + i);
      nbytelen ++;
    }
#else
  for (k = 0; k < s->_nr_pr; k ++)
    j += col_p(k, scratch + j);

  n = ordinal(scratch, j, 2);
  lin [i ++] = n & 255;
  i += nbytes [nbytelen] = bytes_no(n, lin + i);
  nbytelen ++;
#endif /* JOINPROCS */

#ifdef SEPQS
  for (k = 0, l = 0; k < s->_nr_qs;  k ++, l ++)
    {
      for (; l < MAXQ && q_free(l); l ++);

      if (l == MAXQ) {
	printf("%s %d: cannot happen (linearize)\n", __FILE__, __LINE__);
	exit(1);
      }

      n = col_q(l, NULL);
      lin [i ++] = n & 255;
      i += nbytes [nbytelen] = bytes_no(n, lin + i);
      nbytelen ++;
    }
#endif /* SEPQS */

  scratch [j ++] = s->_nr_pr;
  scratch [j ++] = s->_nr_qs;

  memcpy(scratch + j, s->heap, s->_hsz);
  j += s->_hsz;

#ifndef SEPQS
  for (k = 0, l = 0; k < s->_nr_qs; k ++, l ++)
    {
      for (; l < MAXQ && q_offset(l) == -1; l ++);

      if (l == MAXQ) {
	printf("%s %d: cannot happen (linearize)\n", __FILE__, __LINE__);
	exit(1);
      }

      j += col_q(l, scratch + j);
    }
#endif /* SEPQS */
  
  j --;

  for (p = 0, q = 6; p < nbytelen; p ++)
    {
      if (q == 6) {
	q = 0;
	scratch [++ j] = 0;
      }
      else
	q += 2;

      scratch [j] |= (nbytes[p] << q);
    }
  
  j ++;

  for (k = 0; k < WS - 1; k ++)
    scratch [j + k] = 0;

  n = ordinal(scratch, j, 0);
  lin [i ++] = n & 255;
  i += k = bytes_no(n, lin + i);
  lin [i ++] = k;

  for (k = 0; k < WS - 1; k ++)
    lin [i + k] = 0;

  return i;
}

#endif /* COLLAPSE */


/*
 * Wrapper double hashing function
 */

void _double_hash(State * sin)
{
  long n;

  n = linearize(sin);
  d_hash(lin, n);
}


static int generic_hash_store(State *s, struct H_el **tab, hash_t hash)
{
  register struct H_el *old = (struct H_el *)0;
  register struct H_el *tmp;
  struct H_el *ntmp;
  long n;
  int m;
  
  n = linearize(s);

  hash(lin, n);
  tmp = tab[j1];

  if (!tmp) 
    {
      tmp = grab_H_el(lin, n);
      tab[j1] = tmp;
    } 
  else 
    {
      for (;; hcmp ++, old = tmp, tmp = tmp->nxt) 
	{
#ifdef COLLAPSE
	  if (tmp->ln != 0) {
	    if (!tmp->nxt)
	      goto Append;
	    
	    continue;
	  }
#endif /* COLLAPSE */

	  m = memcmp((char *)&tmp->state, lin, n);

	  if (m == 0) 
	    {
#if defined(FULLSTACK)
	      if (tmp->tagged) {
		Lstate = tmp;
		return 2;
	      }
#endif /* defined(FULLSTACK) */	      
	      return 1;
	    }
	  else {
	    if (m < 0)
	      {
		ntmp = grab_H_el(lin, n);
		ntmp->nxt = tmp;

		if (!old)
		  tab[j1] = ntmp;
		else
		  old->nxt = ntmp;
		tmp = ntmp;

		break;
	      }
	    else
	      if (!tmp->nxt) 
		{
#ifdef COLLAPSE
		Append:
#endif
		  tmp->nxt = grab_H_el(lin, n);		  
		  tmp = tmp->nxt;

		  break;
		}
	  }
	}      
    }

#if defined(FULLSTACK)
  tmp->tagged = depth + 1;
  Lstate = tmp;
#endif /* defined(FULLSTACK) */

  return 0;  
}


/*
 * Normal hash store
 */

int _hstore(State *sin)
{
  return generic_hash_store(sin, H_tab, s_hash);
}


#if defined(BITSTATE) && defined(FULLSTACK)

/*
 * Initialize bitstate hash table
 */

void _onstack_init()
{
  S_tab = (struct H_el **) emalloc((1 << (ssize - 3)) * sizeof(struct H_el *));
}


/*
 * Remove the current state from the bitstate hash table and 
 * inserts the element into the hash elements free list.
 */

void _onstack_zap()
{
  register struct H_el * old = (struct H_el *)0;
  register struct H_el * tmp;
  register struct H_el * free;
  long n;
  int m;

  n = linearize(&now);
  b_hash(lin, n);

  for (tmp = S_tab[j1]; tmp; old = tmp, tmp = tmp->nxt)
    {	   
      m = memcmp((char *)&(tmp->state), lin, n);

      if (m == 0)
	goto Found;
      
      if (m < 0)
	break;
    }

  if (!tmp) {
    printf("%s %d: cannot happen (_onstack_zap)\n", __FILE__, __LINE__);
    exit(1);
  }

 Found:
  ZAPS ++;
  
  if (old)
    old->nxt = tmp->nxt;
  else
    S_tab[j1] = tmp->nxt;

  tmp->tagged  = n;

#if !defined(NOREDUCE) && !defined(SAFETY)
  tmp->proviso = 0;
#endif

  tmp->nxt = old = (struct H_el *)0;

  for (free = FreeList; free; old = free, free = free->nxt)
    {
      if (free->tagged <= n)
	{
	  if (old) {
	    tmp->nxt = free->nxt;
	    old->nxt = tmp;
	  }
	  else {
	    tmp->nxt = FreeList;
	    FreeList = tmp;
	  }

	  return;
	}

      if (!tmp->nxt) {
	free->nxt = tmp;
	return;
      }
    }

  FreeList = tmp;
}


/*
 * Inserts the current state into the bitstate hash table.
 */

void _onstack_put()
{
  if (generic_hash_store(&now, S_tab, b_hash) != 0) {
    printf("%s %d: cannot happen (_onstack_put)\n", __FILE__, __LINE__);
    exit(1);
  }

  trpt->ostate = Lstate;
  PUT ++;
}


/*
 * Tells us if the current state is in the hash table
 */

int _onstack_now()
{
  register struct H_el *tmp;
  long n;
  int m = 1;

  n = linearize(&now);
  b_hash(lin, n);

  for (tmp = S_tab[j1]; tmp; tmp = tmp->nxt)
    {
      m = memcmp((char *)&(tmp->state), lin, n);

      if (m <= 0) {
	Lstate = tmp;
	break;
      }
    }

  PROBE ++;

  return (m == 0);
}

#endif /* defined(BITSTATE) && defined(FULLSTACK) */

#ifdef MA

/*
 * Minimized automaton store.
 */

int _gstore(State *s, uchar pbit)
{
  int n = linearize(s);
  static uchar Info[MA+1];

  if (n >= MA)
    {	
      printf("pan: error, MA too small, recompile pan.c");
      printf(" with -DMA=N with N>%d\n", n);
      Uerror("aborting");
    }
  
  if (n > maxgs) 
    maxgs = n;

  memcpy(Info, lin, n);
  memset(Info + n, 0, MA - n - 1);

  Info [MA - 1] = pbit;

  if (a_cycles)	/* place _a_t at the end */
    {	
      Info [MA] = Info [0]; 
      Info [0]	= 0;  
    }

  if (!dfa_store(Info))
    {	
      if ((pbit == 0) && (now._a_t & 1) && (depth > A_depth))
	{	
	  Info[MA] &= ~(1 | 16 | 32);	/* _a_t */
	  if (!dfa_member(MA))
	    return 3;	  
	}

      return 0;	/* new state */
    }

#ifdef FULLSTACK
  if (pbit == 0)
    {	
      Info [MA - 1] = 1;	/* proviso bit */
      trpt->proviso = dfa_member(MA - 1);
      Info [MA - 1] = 4;	/* off-stack bit */
      if (dfa_member(MA - 1))
	return 1; /* off-stack */
      else 
	return 2; /* on-stack */
    }
#endif

  return 1;	/* old state */
}

#endif /* MA */


int endstate(void)
{	
  int i; 
  P *ptr;

  for (i = 0; i <  now._nr_pr; i++)
    {	
      ptr = pptr(i);
      if (!stopstate[ptr->_t][ptr->_p])
	return 0;
    }

  if (strict) 
    return qs_empty();

#if defined(EVENT_TRACE) && !defined(OTIM)
  if (!stopstate[EVENT_TRACE][now._event] && !a_cycles)
    {	
      printf("pan: event_trace not completed\n");
      return 0;
    }
#endif
  return 1;
}


void wrap_stats(void)
{	
#ifdef COVEST
  extern double log(double);
#endif

  printf("%8g states, stored\n", nstates);
  printf("%8g states, matched\n", truncs);
  printf("%8g transitions (= stored+matched)\n", nstates + truncs);
  printf("%8g atomic steps\n", nlinks);

  if (nlost) 
    printf("%g lost messages\n", (double)nlost);

#ifdef BITSTATE
  {
    double a, b;
    
#ifdef CHECK
    printf("%8g states allocated for dfs stack\n", ngrabs);
#endif

#ifndef TEST
    a = (double) (1 << ssize);
    b = nstates + 1.;

#ifdef COVEST
    printf("coverage estimate: %0.1f%%\n",
	   (100. * b) / (log(1. - b / a) / log(1. - 1. / a)));
#endif

    printf("hash factor: %g ", a / b);

    if (!single)
      { 
	if (a / b > 100.)
#ifdef COVEST
	  printf("(good confidence estimate)\n");
	else {
	  if (a / b > 10.)
	    printf("(medium confidence estimate)\n");
	  else
	    printf("(low confidence estimate, best if >100)\n");
	}
      } else
	{ 
	  if (a / b > 1000.)
	    printf("(good confidence estimate)\n");
	  else {
	    if (a / b > 100.)
	      printf("(medium confidence estimate)\n");
	    else
	      printf("(low confidence estimate (1-bit hash), best if>1000)\n");
	  }
#else
	  printf("(expected coverage: >= 99.9%% on avg.)\n");
	  else { 
	    if (a / b > 10.)
	      printf("(expected coverage: >= 98%% on avg.)\n");
	    else
	      printf("(best coverage if >100)\n");
	  }
	} else
	  { 
	    if (a/b > 1000.)
	      printf("(expected coverage: >= 99.9%% on avg.)\n");
	    else {
	      if (a/b > 100.)
		printf("(expected coverage: >= 98%% on avg.)\n");
	      else
		printf("(best coverage (1-bit hash) if >1000)\n");
	    }
#endif /* COVEST */
	}
#endif /* TEST */
  }
#else
#ifndef TEST
  printf("hash conflicts: %g (resolved)\n", hcmp);
#endif /* TEST */
#endif /* BITSTATE */
}


void wrapup(int arg)
{	
#ifdef MEMCNT
#if defined(BITSTATE) || !defined(NOCOMP) 
  double nr1, nr2, nr3 = 0.0, nr4;
#endif
#if !defined(NOCOMP)
  double tmp_nr;
#endif
#endif

  if (spsec && gettimeofday(&stop_tp, NULL)) {
    fprintf(efd, "time get fail: %s\n", strerror(errno));
    exit(0);
  }

  signal(SIGINT, SIG_DFL);
  printf("(%s)\n", Version);

  if (!done) 
    printf("Warning: Search not completed\n");

#ifndef NOREDUCE
  printf("\t+ Partial Order Reduction\n");
#endif

#ifdef POSYMM
  printf("\t+ Symmetry Reduction (Partial)\n");
#endif

#ifdef FULLSYMM
  printf("\t+ Symmetry Reduction (Canonical)\n");
#endif

#ifdef GC0
  printf("\t+ Garbage Collection (Reference Counting)\n");
#endif

#ifdef GC1
  printf("\t+ Garbage Collection (Mark and Sweep)\n");
#endif

#ifdef COLLAPSE
  printf("\t+ Compression\n");
#endif

#ifdef MA
  printf("\t+ Graph Encoding (-DMA=%d)\n", MA);
#ifdef R_XPT
  printf("\t Restarted from checkpoint %s.xpt\n", Source);
#endif
#endif /* MA */

#ifdef CHECK
#ifdef FULLSTACK
  printf("\t+ FullStack Matching\n");
#endif

#ifdef CNTRSTACK
  printf("\t+ CntrStack Matching\n");
#endif
#endif /* CHECK */

#ifdef BITSTATE
  printf("\nBit statespace search for:\n");
#else
  printf("\nFull statespace search for:\n");
#endif

#ifdef EVENT_TRACE
#ifdef NEGATED_TRACE
  printf("	notrace assertion  	+\n");
#else
  printf("	trace assertion    	+\n");
#endif
#endif /* EVENT_TRACE */

#ifdef VERI
  printf("	never-claim         	+\n");
  printf("	assertion violations	");

  if (noasserts)
    printf("- (disabled by -A flag)\n");
  else
    printf("+ (if within scope of claim)\n");
#else
#ifdef NOCLAIM
  printf("	never-claim         	- (not selected)\n");
#else
  printf("	never-claim         	- (none specified)\n");
#endif
  printf("	assertion violations	");
  if (noasserts)
    printf("- (disabled by -A flag)\n");
  else
    printf("+\n");
#endif /* VERI */

#ifndef SAFETY
#ifdef NP
  printf("	non-progress cycles 	");
#else
  printf("	acceptance   cycles 	");
#endif
  if (a_cycles)
    printf("+ (fairness %sabled)\n", fairness ? "en" : "dis");
  else printf("- (not selected)\n");
#else
  printf("	cycle checks       	- (disabled by -DSAFETY)\n");
#endif /* SAFETY */

#ifdef VERI
  printf("	invalid endstates	- ");
  printf("(disabled by ");

  if (noends)
    printf("-E flag)\n\n");
  else
    printf("never-claim)\n\n");
#else
  printf("	invalid endstates	");

  if (noends)
    printf("- (disabled by -E flag)\n\n");
  else
    printf("+\n\n");
#endif /* VERI */

  printf("State-vector %d byte, depth reached %d, errors: %d\n", 
	 hmax, mreached - depth_adjust, errors);

#ifdef MA
  if (done)
    {	
      if (maxgs + a_cycles + 2 < MA)
	printf("MA stats: -DMA=%d is sufficient\n", maxgs + a_cycles + 2);
      dfa_stats();
    }
#endif /* MA */

  wrap_stats();

#ifndef TEST
  printf("(max size 2^%d states)\n\n", ssize);

#ifdef MEMCNT
#if defined(BITSTATE) || !defined(NOCOMP)
  nr1 = nstates * (double)(hmax + sizeof(struct H_el) - sizeof(unsigned));
  nr2 = (double) ((maxdepth + 3) * sizeof(Trail));
#ifndef BITSTATE
#if !defined(MA) || defined(COLLAPSE)
  nr3 = (double) (1 << ssize) * sizeof(struct H_el *);
#endif
#else /* BITSTATE */
  nr3 = (double) (1 << (ssize - 3));
#ifdef CNTRSTACK
  nr3 += (double) (1 << (ssize - 3));
#endif 
#endif /* BITSTATE */

  nr4 = (double) (svmax * (sizeof(Svtack) + hmax)) 
    + (double) (smax * (sizeof(Stack)));

#ifndef MA
  if (memcnt < nr1 + nr2 + nr3 + nr4)
#else
    if (1)
#endif
      {	
	printf("Stats on memory usage (in Megabytes):\n");
	printf("%-6.3f	equivalent memory usage for states", nr1 / 1000000.);
	printf(" (stored * (State-vector))\n");
#ifdef BITSTATE
	printf("%-6.3f	memory used for hash-array (-w%d)\n", nr3 / 1000000., 
	       ssize);
#else
	tmp_nr = memcnt - nr3 - nr4 - nr2;

	if (tmp_nr < 0.0) 
	  tmp_nr = 0.;
	
	printf("%-6.3f	actual memory usage for states", tmp_nr / 1000000.);
	printf(" (");
	    
	if (tmp_nr > 0.)
	  {	
	    if (tmp_nr > nr1)
	      printf("unsuccessful ");

	    printf("compression: %.2f%%)\n", (100.0*tmp_nr)/nr1);
	  } else
	    printf("less than 1k)\n");
	    
#ifndef MA
	if (tmp_nr > 0.)
	  {	
	    printf("	State-vector as stored = %.0f byte",
		   (tmp_nr) / nstates - 
		   (double)(sizeof(struct H_el) - sizeof(unsigned)));
	      
	    printf(" + %d byte overhead\n",
		   sizeof(struct H_el) - sizeof(unsigned));
	  }
#endif

#if !defined(MA) || defined(COLLAPSE)
	printf("%-6.3f	memory used for hash-table (-w%d)\n", 
	       nr3 / 1000000., ssize);
#endif	  
#endif /* defined(BITSTATE) || !defined(NOCOMP) */

	printf("%-6.3f	memory used for DFS stack (-m%ld)\n", 
	       nr2 / 1000000., maxdepth);

	/* remainder is mem used for proc and chan stacks */
	/* and memory lost in allocator (=fragment) */
	printf("%-6.3f	total actual memory usage\n\n", memcnt / 1000000.);
      } else
#endif
	printf("%-6.3f	memory usage (Mbyte)\n", memcnt / 1000000.);
#endif /* MEMCNT */

#ifdef COLLAPSE
  printf("nr of templates: [ globals procs chans ]\n");
  printf("collapse counts: [ ");
  { 
    int i; 
    
    for (i = 0; i < 256 + 2; i ++)
      if (ncomps[i] != 0)
	printf("%ld ", ncomps[i]);

    printf("]\n");
  }
#endif /* COLLAPSE */
#endif /* TEST */

#ifndef NOGC
  printf("\n  + %ld ", gc_runs);
#ifdef GC0
  printf("reference counting collector runs\n\n");
#endif
#ifdef GC1
  printf("mark and sweep collector runs\n");

  if (gc_runs)
    printf("  + average collection rate: %-2.1f%%\n\n", 
	   gc_coef / (double) gc_runs * 100);
  else
    printf("\n");
#endif
#endif /* NOGC */

  if (done && !no_rck) 
    do_reach();

#ifdef PEG
  { 
    int i;
    printf("\nPeg Counts (transitions executed):\n");
    
    for (i = 1; i < NTRANS; i++)
      {	
	if (peg[i]) 
	  putpeg(i, peg[i]);
      } 
  }
#endif /* PEG */

#ifdef VAR_RANGES
  dumpranges();
#endif /* VAR_RANGES */

#ifdef SVDUMP
  if (vprefix > 0) 
    close(svfd);
#endif /* SVDUMP */

  if (spsec)
    {
      double delta = stop_tp.tv_sec - start_tp.tv_sec;
      printf("\nestimated speed: ");
      if (delta)
	printf("%8g states/sec\n", nstates / delta);
      else
	printf("time slice too small\n");
    }
  
  exit(0);
}


static void xrefsrc(int lno, int M, int i)
{	
  Trans *T;

  for (T = trans[M][i]; T; T = T->nxt)
    if (T && T->tp)
      {	
	printf("\tline %d, state %d", lno, i);

	if (strcmp(T->tp, "") != 0)
	  printf(", \"%s\"", T->tp);
	else if (stopstate[M][i])
	  printf(", -endstate-");

	printf("\n");
      }
}


void r_ck(uchar *which, int N, int M, short *src)
{	
  int i, m = 0;

#ifdef VERI
  if (M == VERI) 
    return;	/* no useful info there */
#endif

  printf("unreached in proctype %s\n", procname [M]);

  for (i = 1; i < N; i++)
    if (which[i] == 0 && trans[M][i])
      xrefsrc(src[i], M, i);
    else
      m++;

  printf("	(%d of %d states)\n", N-1-m, N-1);
}


void putrail(void)
{	
  int fd, i, j, k;
  char snap[64], fnm[256];
  char *skip = (char *) emalloc(depth);

  if (iterative == 0 && Nr_Trails ++ > 0)
    sprintf(fnm, "%s%d.trail", Source, Nr_Trails);
  else
    sprintf(fnm, "%s.trail", Source);

  if ((fd = creat(fnm, 0666)) <= 0)
    {	
      printf("cannot create %s\n", fnm);
      return;
    }

#ifdef VERI
  sprintf(snap, "-2:%d:-2\n", VERI);
  write(fd, snap, strlen(snap));
#endif

  for (i = k = 1; i <= depth; i ++)
    {	
      if (i == depthfound + 1)
	write(fd, "-1:-1:-1\n", 9);

      if (trail[i].o_pm & 128 || skip[i]) 
	continue;

      sprintf(snap, "%d:%d:%d\n", k ++, trail[i].pr, trail[i].o_t->t_id);

      j = strlen(snap);

      if (write(fd, snap, j) != j)
	{	
	  printf("pan: error writing %s\n", fnm);
	  close(fd);
	  wrapup(0);
	}

      /* skip the next move of the same process */      
      if (trail[i].rt) 
	{
	  int l;
	  
	  for (l = i + 1; l <= depth; l ++)
	    if (trail[l].pr == trail[i].pr)
	      {
		skip[l] = 1;
		break;
	      }
	}    
    }
  
  printf("pan: wrote %s\n", fnm);
  efree(skip, depth);
  close(fd);
}


void snapshot()
{
  printf("Depth= %7d States= %7g ", mreached, nstates);
  printf("Transitions= %7g ", nstates+truncs);

#ifdef MA
  printf("Nodes= %7ld ", nr_states);
#endif

  printf("Memory= %-6.3f\n", memcnt/1000000.);
  fflush(stdout);
}


void Uerror(char *str)
{	
  /* always fatal */
  uerror(str);
  wrapup(0);
}


void fatal(char *str)
{
  /* no trail dump */
  printf(str);
  wrapup(0);
}


#if defined(MA) && !defined(SAFETY)
int Unwind(void)
{	
  Trans *t; 
  char ot, m; 
  short tt; 
  short II;
  State tmp;

  uchar oat = now._a_t;
  now._a_t &= ~(1 | 16 | 32);
  _statecpy(&tmp, &now);
  now._a_t = oat;

Up:
#ifndef NOFAIR
  if (trpt->o_pm & 128)	/* fairness alg */
    {	
      now._cnt [now._a_t & 1] = trpt->oval;
      depth --; 
      trpt  --;
      goto Q999;
    }
#endif

#ifdef HAS_LAST
#ifdef VERI
  { 
    int d;
    now._last = 0;

    for (d = 1; d < depth; d ++)
      {	
	if ((trpt - d)->pr != 0)
	  { 
	    now._last = (trpt-d)->pr - BASE;
	    break;
	  } 
      }	
  }
#else
  now._last = (depth < 1) ? 0 : (trpt - 1) -> pr;
#endif
#endif /* HAS_LAST */

#ifdef EVENT_TRACE
  now._event = trpt->o_event;
#endif

  if ((now._a_t & 1) && depth <= A_depth)
    {	
      now._a_t &= ~(1 | 16 | 32);
      if (fairness) 
	now._a_t |= 2;	

      A_depth = 0;
      goto CameFromHere;	/* checkcycles() */
    }
	
  t  = trpt->o_t;
  ot = trpt->o_ot; 
  II = trpt->pr;
  tt = trpt->o_tt; 
  this = pptr(II);

  vt[II] = trpt->vt;
  rt[II] = trpt->rt;

  m = do_reverse(t, II, trpt->o_m);

  depth --; 
  trpt  --;

  reached[(int)ot][t->st] = 1;
  this->_p = tt;

#ifndef NOFAIR
  if ((trpt->o_pm & 32))
    {
#ifdef VERI
      if (now._cnt[now._a_t & 1] == 0)
	now._cnt[now._a_t & 1] = 1;
#endif
      now._cnt[now._a_t & 1] += 1;
    }

Q999:
  if (trpt->o_pm & 8)
    {	
      now._a_t &= ~2;
      now._cnt[now._a_t & 1] = 0;
    }
	
  if (trpt->o_pm & 16)
    now._a_t |= 2;
#endif /* NOFAIR */

CameFromHere:
  if (_statecmp(&now, &tmp) == 0)
    return depth;

  if (depth > 0) 
    goto Up;

  return 0;
}

#endif /* defined(MA) && !defined(SAFETY) */


void uerror(char *str)
{	
  static char laststr[256];

  if (strcmp(str, laststr))
    printf("pan: %s (at depth %ld)\n", str,
	   (depthfound == -1 ? depth : depthfound) - depth_adjust);

  strcpy(laststr, str);
  errors++;

  if (every_error != 0 ||  errors == upto)
    {
#if defined(MA) && !defined(SAFETY)
      if (strstr(str, " cycle"))
	{	
	  int od = depth;
	  depthfound = Unwind();
	  depth = od;
	}
#endif

      putrail();

#if defined(MA) && !defined(SAFETY)
      if (strstr(str, " cycle"))
	{	
	  if (every_error)
	    printf("sorry: MA writes 1 trail max\n");

	  wrapup(0); /* no recovery from unwind */
	}
#endif
    }
  
  if (iterative != 0 && maxdepth > 0)
    {	
      maxdepth = (iterative == 1) ? (depth-1) : (depth / 2);
      warned = 1;
      printf("pan: reducing search depth to %ld\n", maxdepth);
    } 
  else 
    if (errors >= upto && upto != 0)
      wrapup(0);

  depthfound = -1;	/* tripakis */
}


void usage(FILE *fd)
{
  fprintf(fd, "Valid Options are:\n");
#ifndef SAFETY
#ifdef NP
  fprintf(fd, "	-a  -> is disabled by -DNP ");
  fprintf(fd, "(-DNP compiles for -l only)\n");
#else
  fprintf(fd, "	-a  find acceptance cycles\n");
#endif
#else
  fprintf(fd, "	-a,-l,-f  -> are disabled by -DSAFETY\n");
#endif
  fprintf(fd, "	-A  ignore assert() violations\n");
  fprintf(fd, "	-cN stop at Nth error ");
  fprintf(fd, "(defaults to -c1)\n");
  fprintf(fd, "	-d  print state tables and stop\n");
  fprintf(fd, "	-e  create trails for all errors\n");
  fprintf(fd, "	-E  ignore invalid endstates\n");
#ifndef NOFAIR
  fprintf(fd, "	-f  add weak fairness (to -a or -l)\n");
#endif
  fprintf(fd, "	-hN choose other hash-function 1..32\n");
  fprintf(fd, "	-i  search for shortest path to error\n");
  fprintf(fd, "	-I  like -i, but approximate and faster\n");
#ifndef SAFETY
#ifdef NP
  fprintf(fd, "	-l  find non-progress cycles\n");
#else
  fprintf(fd, "	-l  find non-progress cycles -> ");
  fprintf(fd, "disabled, requires ");
  fprintf(fd, "compilation with -DNP\n");
#endif
#endif
  fprintf(fd, "	-mN max depth N steps (default=10k)\n");
  fprintf(fd, "	-n  no listing of unreached states\n");
#ifdef SVDUMP
  fprintf(fd, "	-pN create svfile (save N bytes per state)\n");
#endif
  fprintf(fd, "	-q  require empty chans in valid endstates\n");
#ifdef BITSTATE
  fprintf(fd, "	-RN repeat run Nx with N ");
  fprintf(fd, "[1..32] independent hash functions\n");
#endif
  fprintf(fd, "	-s  1-bit hashing (default is 2-bit)\n");
  fprintf(fd, "	-t  print estimated speed\n");
  fprintf(fd, "	-V  print SPIN version number\n");
  fprintf(fd, "	-wN hashtable of 2^N entries");
  fprintf(fd, "(defaults to -w%d)\n", ssize);
  exit(1);
}


void to_compile(void)
{	
  char ctd[1024], carg[64]; 
  int n;

#ifdef BITSTATE
  strcpy(ctd, "-DBITSTATE ");
#else
  strcpy(ctd, "");
#endif

#ifdef PC
  strcat(ctd, "-DPC ");
#endif

#ifdef NOVSZ
  strcat(ctd, "-DNOVSZ ");
#endif

#ifdef MEMCNT
#ifdef PC
  n = 25;
#else
  n = 28;
#endif
  if (MEMCNT != n)
    {	
      sprintf(carg, "-DMEMCNT=%d ", MEMCNT);
      strcat(ctd, carg);
    }
#endif

#ifdef NOCLAIM
  strcat(ctd, "-DNOCLAIM ");
#endif

#ifdef SAFETY
  strcat(ctd, "-DSAFETY ");
#else

#ifdef NOFAIR
  strcat(ctd, "-DNOFAIR ");
#else

#ifdef NFAIR
  if (NFAIR != 2)
    {	
      sprintf(carg, "-DNFAIR=%d ", NFAIR);
      strcat(ctd, carg);
    }
#endif
#endif
#endif

#ifdef NOREDUCE
  strcat(ctd, "-DNOREDUCE ");
#else

#ifdef XUSAFE
  strcat(ctd, "-DXUSAFE ");
#endif
#endif

#ifdef NP
  strcat(ctd, "-DNP ");
#endif

#ifdef PEG
  strcat(ctd, "-DPEG ");
#endif

#ifdef VAR_RANGES
  strcat(ctd, "-DVAR_RANGES ");
#endif

#ifdef CHECK
  strcat(ctd, "-DCHECK ");
#endif

#ifdef CTL
  strcat(ctd, "-DCTL ");
#endif

#ifdef NIBIS
  strcat(ctd, "-DNIBIS ");
#endif

#ifdef NOBOUNDCHECK
  strcat(ctd, "-DNOBOUNDCHECK ");
#endif

#ifdef NOSTUTTER
  strcat(ctd, "-DNOSTUTTER ");
#endif

#ifdef REACH
  strcat(ctd, "-DREACH ");
#endif

#ifdef PRINTF
  strcat(ctd, "-DPRINTF ");
#endif

#ifdef OTIM
  strcat(ctd, "-DOTIM ");
#endif

#ifdef COLLAPSE
  strcat(ctd, "-DCOLLAPSE ");
#endif

#ifdef SVDUMP
  strcat(ctd, "-DSVDUMP ");
#endif

#ifdef VERBOSE
  strcat(ctd, "-DVERBOSE ");
#endif

#ifdef CHECK
  strcat(ctd, "-DCHECK ");
#endif

#ifdef SDUMP
  strcat(ctd, "-DSDUMP ");
#endif

#ifdef COVEST
  strcat(ctd, "-DCOVEST ");
#endif

#ifdef MA
  strcat(ctd, "-DMA ");
#endif
 
  printf("Compiled as: cc -o pan %span.c", ctd);

#if defined(COVEST) && defined(BITSTATE)
  printf(" -lm\n");
#else
  printf("\n");
#endif
}


int main(int argc, char *argv[])
{	
#if defined(SAFETY) && defined(NP)
  fprintf(efd, "error: cannot combine -DNP and -DSAFETY\n");
  exit(1);
#endif

#ifdef MA
#ifdef BITSTATE
  fprintf(efd, "error: cannot combine -DMA and -DBITSTATE\n");
  exit(1);
#endif /* BITSTATE */
  if (MA <= 0)
    {	
      fprintf(efd, "usage: -DMA=N with N > 0 and < VECTORSZ\n");
      exit(1);
    }
#endif /* MA */

#ifdef COLLAPSE
#if defined(BITSTATE)
  fprintf(efd, "error: cannot combine -DBITSTATE and -DCOLLAPSE\n");
  exit(1);
#endif

#if defined(NOCOMP)
  fprintf(efd, "error: cannot combine -DNOCOMP and -DCOLLAPSE\n");
  exit(1);
#endif
#endif /* NOCOMP */

#if defined(NOGC) && (defined(GC0) || defined(GC1))
  fprintf(efd, "error: -DNOGC cannot be combined with explicit collection\n");
  exit(1);
#endif /* NOGC && (GC0 || GC1) */

#if defined(NOSYMM) && (defined(POSYMM) || defined(FULLSYMM))
  fprintf(efd, "error: -DNOSYMM cannot be combined with symmetry reduction\n");
  exit(1);
#endif /* NOSYMM && FULLSYMM */

#if !defined(NOSYMM) && defined(GC0)
  fprintf(efd, "warning: -DGC0 automatically disables symmetry reductions\n");
#endif /* !NOSYMM && GC0 */

#if !defined(NOSYMM) && defined(NOGC)
  fprintf(efd, "warning: -DNOGC automatically disables symmetry reductions\n");
#endif /* !NOSYMM && NOGC */

#if defined(GC0) && defined(GC1)
  fprintf(efd, "error: cannot combine -DGC0 and -DGC1\n");
  exit(1);
#endif

#if defined(POSYMM) && defined(FULLSYMM)
  fprintf(efd, "error: cannot combine -DPOSYMM and -DFULLSYMM\n");
  exit(1);
#endif

  while (argc > 1 && argv[1][0] == '-')
    {	
      switch (argv[1][1]) {
#ifndef SAFETY
#ifdef NP
      case 'a': 
	fprintf(efd, "error: -a disabled");
	usage(efd); 
	break;
#else
      case 'a': 
	a_cycles = 1; 
	break;
#endif /* NP */
#endif /* SAFETY */
      case 'A': 
	noasserts = 1; 
	break;

      case 'c': 
	upto  = atoi(&argv[1][2]); 
	break;

      case 'd': 
	state_tables = 1; 
	break;

      case 'e': 
	every_error = 1; 
	break;

      case 'E': 
	noends = 1; 
	break;
#if !defined(SAFETY) && !defined(NOFAIR)
      case 'f': 
	fairness = 1; 
	break;
#endif /* !defined(SAFETY) && !defined(NOFAIR) */

      case 'h': 
	if (!argv[1][2]) 
	  usage(efd); 
	else
	  HASH_NR = atoi(&argv[1][2]) % 33; 
	break;

      case 'I': 
	iterative = 2; 
	every_error = 1; 
	break;

      case 'i': 
	iterative = 1; 
	every_error = 1; 
	break;

#ifndef SAFETY
#ifdef  NP
      case 'l': 
	a_cycles = 1; 
	break;
#else
      case 'l': 
	fprintf(efd, "error: -l disabled");
	usage(efd); 
	break;
#endif /* NP */
#endif /* SAFETY */

      case 'm': 
	maxdepth = atoi(&argv[1][2]); 
	break;

      case 'n': 
	no_rck = 1; 
	break;

#ifdef SVDUMP
      case 'p': 
	vprefix = atoi(&argv[1][2]); 
	break;
#endif
      case 'q': 
	strict = 1; 
	break;

      case 'R': 
	Nrun = atoi(&argv[1][2]); 
	break;

      case 's': 
	single = 1; 
	break;

      case 't':
	spsec = 1;
	break;

      case 'V': 
	printf("Generated by %s\n", Version);
	to_compile(); 
	exit(0); 
	break;

      case 'w': 
	ssize = atoi(&argv[1][2]); 
	break;

      default : 
	usage(efd); 
	break;
      }

      argc--; 
      argv++;
    }

  if (maxdepth <= 0 || ssize <= 0) 
    usage(efd);

#if SYNC>0 && !defined(NOREDUCE)
  if (a_cycles && fairness)
    { 
      fprintf(efd, "error: p.o. reduction not compatible with ");
      fprintf(efd, "fairness (-f) in models\n");
      fprintf(efd, "       with rendezvous operations: ");
      fprintf(efd, "recompile with -DNOREDUCE\n");
      exit(1);
    }
#endif

#if defined(NOCOMP) && !defined(BITSTATE)
  if (a_cycles)
    { 
      fprintf(efd, "error: -DNOCOMP voids -l and -a\n");
      exit(1);
    }
#endif

#ifndef BITSTATE
  if (Nrun > 1) 
    HASH_NR = Nrun - 1;
#endif

  if (Nrun < 1 || Nrun > 32)
    {	
      fprintf(efd, "error: invalid arg for -R\n");
      usage(efd);
    }

#ifndef SAFETY
  if (fairness && !a_cycles)
    {	
      fprintf(efd, "error: -f requires -a or -l\n");
      usage(efd);
    }

#if ACCEPT_LAB == 0
  if (a_cycles)
#ifndef VERI
    {	
      fprintf(efd, "error: no accept labels defined ");
      fprintf(efd, "in model (for option -a)\n");
      usage(efd);
    }
#else
    {	
      fprintf(efd, "warning: no explicit accept labels ");
      fprintf(efd, "defined in model (for -a)\n");
    }
#endif /* VERI */
#endif /* ACCEPT_LAB == 0 */
#endif /* SAFETY */

#if !defined(NOREDUCE)
#if defined(HAS_ENABLED)
  fprintf(efd, "error: reduced search precludes ");
  fprintf(efd, "use of 'enabled()'\n");
  exit(1);
#endif

#if defined(HAS_PCVALUE)
  fprintf(efd, "error: reduced search precludes ");
  fprintf(efd, "use of 'pcvalue()'\n");
  exit(1);
#endif
#if defined(HAS_BADELSE)
  fprintf(efd, "error: reduced search precludes ");
  fprintf(efd, "using 'else' combined with i/o stmnts\n");
  exit(1);
#endif
#if defined(HAS_LAST)
  fprintf(efd, "error: reduced search precludes ");
  fprintf(efd, "use of _last\n");
  exit(1);
#endif
#endif /* defined(HAS_PCVALUE) */

#if !defined(REACH) && !defined(BITSTATE)
  if (iterative != 0)
    fprintf(efd, "warning: -i needs -DREACH or -DBITSTATE\n");
#endif

#if defined(BITSTATE) && defined(REACH)
  fprintf(efd, "warning: -DREACH voided by -DBITSTATE\n");
#endif

#if defined(FULLSTACK) && defined(CNTRSTACK)
  fprintf(efd, "error: cannot combine");
  fprintf(efd, " -DFULLSTACK and -DCNTRSTACK\n");
  exit(1);
#endif

#ifdef VERI
#if ACCEPT_LAB > 0
  if (!a_cycles && !state_tables)
    { 
      fprintf(efd, "warning: never-claim + accept-labels ");
      fprintf(efd, "requires -a flag to fully verify\n");
    }
#endif
#endif /* VERI */

#ifndef SAFETY
  if (!a_cycles && !state_tables)
    { 
      fprintf(efd, "hint: this search is more efficient ");
      fprintf(efd, "if pan.c is compiled -DSAFETY\n");
    }
#endif /* SAFETY */

  signal(SIGINT, wrapup);
  mask = (1 << ssize) - 1;
  trail = (Trail *) emalloc((maxdepth+4)*sizeof(Trail));
  trail ++;	/* protect trpt-1 refs at depth 0 */

#ifdef RANDSTOR
  srand(123);
#endif

  run();

  done = 1;
  wrapup(0);

  /* just to get rid of the warning */
  return 0;
}


static int addr_in_range(long x)
{
  register int i;
  register P  *p;

  if (((long)now.heap <= x) && (x < (long)now.heap + now._hsz))
    return 1;

  for (i = 0; i < now._nr_pr; i ++)
    {
      p = pptr(i);

      if (((long)p <= x) && (x < (long)p + p_sz(p)))
	return 1;
    }

  return 0;
}


inline char *xptr(char *x)
{
  if (!addr_in_range((long)x))
    Uerror("address out of range");

  return x;
}


void struct_copy(char *d, char *s, long sz)
{
  if (!addr_in_range((long)s))
    Uerror("address out of range");

  memcpy(d, s, sz);
}


#ifdef MA

/* The DFA code below was written by Anuj Puri and Gerard J. Holzmann in  */
/* May 1997, and was inspired by earlier work on data compression using   */
/* sharing tree data structures and graph-encoded sets by J-Ch. Gregoire  */
/* (INRS Telecom, Quebec, Canada) and D.Zampunieris (Univ.Namur, Belgium) */


#define ulong	unsigned long
#define ushort	unsigned short

#define TWIDTH		256
#define HASH(y,n)	(n)*(((int)y))
#define INRANGE(e,h)	((h>=e->From && h<=e->To)||(e->s==1 && e->S==h))

typedef struct Edge {
	uchar From, To;		/* max range 0..255 */
	uchar s, S;		/* if s=1, S is singleton */
	struct Vertex	*Dst;
	struct Edge	*Nxt;
} Edge;

typedef struct Vertex {
	ulong	key, num;      /* key for splay tree, nr incoming edges */
	uchar	from[2], to[2];	/* in-node predefined edge info    */
	struct	Vertex	*dst[2];/* most nodes have 2 or more edges */
	struct	Edge	*Succ;	/* in case there are more edges */
	struct	Vertex	*lnk, *left, *right; /* splay tree plumbing */
} Vertex;

static Edge	*free_edges;
static Vertex	*free_vertices;
static Vertex	**layers;	/* one splay tree of nodes per layer */
static Vertex	**path;		/* run of word in the DFA */
static Vertex	*R, *F, *NF;	/* Root, Final, Not-Final */
static uchar	*word, *lastword;/* string, and last string inserted */
static int	dfa_depth, iv=0, nv=0, pfrst=0, Tally;

static void	insert_it(Vertex *, int);	/* splay-tree code */
static void	delete_it(Vertex *, int);
static Vertex	*find_it(Vertex *, Vertex *, uchar, int);

static void
recyc_edges(Edge *e)
{
	if (!e) return;
	recyc_edges(e->Nxt);
	e->Nxt = free_edges;
	free_edges = e;
}

static Edge *
new_edge(Vertex *dst)
{	Edge *e;

	if (free_edges)
	{	e = free_edges;
		free_edges = e->Nxt;
		e->From = e->To = e->s = e->S = 0;
		e->Nxt = (Edge *) 0;
	} else
		e = (Edge *) emalloc(sizeof(Edge));
	e->Dst = dst;

	return e;
}

static void
recyc_vertex(Vertex *v)
{
	recyc_edges(v->Succ);
	v->Succ = (Edge *) free_vertices;
	free_vertices = v;
	nr_states--;
}

static Vertex *
new_vertex(void)
{	Vertex *v;

	if (free_vertices)
	{	v = free_vertices;
		free_vertices = (Vertex *) v->Succ;
		v->Succ = (Edge *) 0;
		v->num  = 0;
	} else
		v = (Vertex *) emalloc(sizeof(Vertex));

	nr_states++;
	return v; 
}

static Vertex *
allDelta(Vertex *v, int n)
{	Vertex *dst = new_vertex();

	v->from[0] = 0;
	v->to[0] = 255;
	v->dst[0] = dst;
	dst->num = 256;
	insert_it(v, n);
	return dst;
}

static void
insert_edge(Vertex *v, Edge *e)
{	/* put new edge first */
	if (!v->dst[0])
	{	v->dst[0] = e->Dst;
		v->from[0] = e->From;
		v->to[0] = e->To;
		recyc_edges(e);
		return;
	}
	if (!v->dst[1])
	{	v->from[1] = v->from[0]; v->from[0] = e->From;
		v->to[1]   = v->to[0];   v->to[0]   = e->To;
		v->dst[1]  = v->dst[0];  v->dst[0]  = e->Dst;
		recyc_edges(e);
		return;
	} /* shift */
	{	int f      = v->from[1];
		int t      = v->to[1];
		Vertex *d  = v->dst[1];
		v->from[1] = v->from[0]; v->from[0] = e->From;
		v->to[1]   = v->to[0];   v->to[0]   = e->To;
		v->dst[1]  = v->dst[0];  v->dst[0]  = e->Dst;
		e->From = f;
		e->To   = t;
		e->Dst  = d;
	}
	e->Nxt = v->Succ;
	v->Succ = e;
}

static void
copyRecursive(Vertex *v, Edge *e)
{	Edge *f;
	if (e->Nxt) copyRecursive(v, e->Nxt);
	f = new_edge(e->Dst);
	f->From = e->From;
	f->To   = e->To;
	f->s    = e->s;
	f->S    = e->S;
	f->Nxt  = v->Succ;
	v->Succ = f;
}

static void
copyEdges(Vertex *to, Vertex *from)
{	int i;
	for (i = 0; i < 2; i++)
	{	to->from[i] = from->from[i];
		to->to[i]   = from->to[i];
		to->dst[i]  = from->dst[i];
	}
	if (from->Succ) copyRecursive(to, from->Succ);
}

static Edge *
cacheDelta(Vertex *v, int h, int first)
{	static Edge *ov, tmp;  int i;

	if (!first && INRANGE(ov,h))
		return ov; /* intercepts about 10% */
	for (i = 0; i < 2; i++)
		if (v->dst[i] && h >= v->from[i] && h <= v->to[i])
		{	tmp.From = v->from[i];
			tmp.To   = v->to[i];
			tmp.Dst  = v->dst[i];
			tmp.s    =  tmp.S = 0;
			ov = &tmp;
			return ov;
		}
	for (ov = v->Succ; ov; ov = ov->Nxt)
		if (INRANGE(ov,h)) return ov;

	Uerror("cannot get here, cacheDelta");
	return (Edge *) 0;
}

static Vertex *
Delta(Vertex *v, int h)	/* v->delta[h] */
{	register Edge *e;

	if (v->dst[0] && h >= v->from[0] && h <= v->to[0])
		return v->dst[0];	/* oldest edge */
	if (v->dst[1] && h >= v->from[1] && h <= v->to[1])
		return v->dst[1];
	for (e = v->Succ; e; e = e->Nxt)
		if (INRANGE(e,h))
			return e->Dst;
	Uerror("cannot happen Delta");
	return (Vertex *) 0;
}

static void
numDelta(Vertex *v, int d)
{	register Edge *e;
	register ulong cnt;
	int i;

	for (i = 0; i < 2; i++)
	if (v->dst[i])
	{	cnt = v->dst[i]->num + d*(1 + v->to[i] - v->from[i]);
		if (d == 1 && cnt < v->dst[i]->num) goto bad;
		v->dst[i]->num = cnt;
	}
	for (e = v->Succ; e; e = e->Nxt)
	{	cnt = e->Dst->num + d*(1 + e->To - e->From + e->s);
		if (d == 1 && cnt < e->Dst->num)
bad:			Uerror("too many incoming edges");
		e->Dst->num = cnt;
	}
}

static void
setDelta(Vertex *v, int h, Vertex *newdst)	/* v->delta[h] = newdst; */
{	Edge *e, *f = (Edge *) 0, *g;
	int i;

	/* remove the old entry, if there */
	for (i = 0; i < 2; i++)
		if (v->dst[i] && h >= v->from[i] && h <= v->to[i])
		{	if (h == v->from[i])
			{	if (h == v->to[i])
				{	v->dst[i] = (Vertex *) 0;
					v->from[i] = v->to[i] = 0;
				} else
					v->from[i]++;
			} else if (h == v->to[i])
			{	v->to[i]--;
			} else
			{	g = new_edge(v->dst[i]);/* same dst */
				g->From    = v->from[i];
				g->To      = h-1;	/* left half */
				v->from[i] = h+1;	/* right half */
				insert_edge(v, g);
			}
			goto part2;
		}
	for (e = v->Succ; e; f = e, e = e->Nxt)
	{	if (e->s == 1 && e->S == h)
		{	e->s = e->S = 0;
			goto rem_tst;
		}
		if (h >= e->From && h <= e->To)
		{	if (h == e->From)
			{	if (h == e->To)
				{	if (e->s)
					{	e->From = e->To = e->S;
						e->s = 0;
						break;
					} else
						goto rem_do;
				} else
					e->From++;
			} else if (h == e->To)
			{	e->To--;
			} else				/* split */
			{	g = new_edge(e->Dst);	/* same dst */
				g->From = e->From;
				g->To   = h-1;		/* g=left half */
				e->From = h+1;		/* e=right half */
				g->Nxt  = e->Nxt;	/* insert g */
				e->Nxt  = g;		/* behind e */
				break;			/* done */
			}

rem_tst:		if (e->From > e->To)
			{	if (e->s == 0) {
rem_do:				if (f)
						f->Nxt = e->Nxt;
					else
						v->Succ = e->Nxt;
					e->Nxt = (Edge *) 0;
					recyc_edges(e);
				} else
				{	e->From = e->To = e->S;
					e->s = 0;
			}	}
			break;
	}	}
part2:
	/* check if newdst is already there */
	for (i = 0; i < 2; i++)
		if (v->dst[i] == newdst)
		{	if (h+1 == (int) v->from[i])
			{	v->from[i] = h;
				return;
			}
			if (h == (int) v->to[i]+1)
			{	v->to[i] = h;
				return;
		}	}
	for (e = v->Succ; e; e = e->Nxt)
	{	if (e->Dst == newdst)
		{	if (h+1 == (int) e->From)
			{	e->From = h;
				if (e->s == 1 && e->S+1 == e->From)
				{	e->From = e->S;
					e->s = e->S = 0;
				}
				return;
			}
			if (h == (int) e->To+1)
			{	e->To = h;
				if (e->s == 1 && e->S == e->To+1)
				{	e->To = e->S;
					e->s = e->S = 0;
				}
				return;
			}
			if (e->s == 0)
			{	e->s = 1;
				e->S = h;
				return;
	}	}	}
	/* add as a new edge */
	e = new_edge(newdst);
	e->From = e->To = h;
	insert_edge(v, e);
}

static ulong
cheap_key(Vertex *v)
{	ulong vk2 = 0;

	if (v->dst[0])
	{	vk2 = (ulong) v->dst[0];
		if ((ulong) v->dst[1] > vk2)
			vk2 = (ulong) v->dst[1];
	} else if (v->dst[1])
		vk2 = (ulong) v->dst[1]; 
	if (v->Succ)
	{	Edge *e;
		for (e = v->Succ; e; e = e->Nxt)
			if ((ulong) e->Dst > vk2)
				vk2 = (ulong) e->Dst;
	}
	Tally = (vk2>>2)&(TWIDTH-1);
	return v->key;
}

static ulong
mk_key(Vertex *v)	/* not sensitive to order */
{	register ulong m = 0, vk2 = 0;
	register Edge *e;

	if (v->dst[0])
	{	m += HASH(v->dst[0], v->to[0] - v->from[0] + 1);
		vk2 = (ulong) v->dst[0]; 
	}
	if (v->dst[1])
	{	m += HASH(v->dst[1], v->to[1] - v->from[1] + 1);
		if ((ulong) v->dst[1] > vk2) vk2 = (ulong) v->dst[1]; 
	}
	for (e = v->Succ; e; e = e->Nxt)
	{	m += HASH(e->Dst, e->To - e->From + 1 + e->s);
		if ((ulong) e->Dst > vk2) vk2 = (ulong) e->Dst; 
	}
	Tally = (vk2>>2)&(TWIDTH-1);
	return m;
}

static ulong
mk_special(int sigma, Vertex *n, Vertex *v)
{	register ulong m = 0, vk2 = 0;
	register Edge *f; 
	int i;

	for (i = 0; i < 2; i++)
		if (v->dst[i])
		{	if (sigma >= v->from[i] && sigma <= v->to[i])
			{	m += HASH(v->dst[i], v->to[i]-v->from[i]);
				if ((ulong) v->dst[i] > vk2
				&&   v->to[i] > v->from[i])
					vk2 = (ulong) v->dst[i]; 
			} else
			{	m += HASH(v->dst[i], v->to[i]-v->from[i]+1);
				if ((ulong) v->dst[i] > vk2)
					vk2 = (ulong) v->dst[i]; 
		}	}
	for (f = v->Succ; f; f = f->Nxt)
	{	if (sigma >= f->From && sigma <= f->To)
		{	m += HASH(f->Dst, f->To - f->From + f->s);
			if ((ulong) f->Dst > vk2
			&&   f->To - f->From + f->s > 0)
				vk2 = (ulong) f->Dst; 
		} else if (f->s == 1 && sigma == f->S)
		{	m += HASH(f->Dst, f->To - f->From + 1);
			if ((ulong) f->Dst > vk2) vk2 = (ulong) f->Dst; 
		} else
		{	m += HASH(f->Dst, f->To - f->From + 1 + f->s);
			if ((ulong) f->Dst > vk2) vk2 = (ulong) f->Dst; 
	}	}

	if ((ulong) n > vk2) vk2 = (ulong) n; 
	Tally = (vk2>>2)&(TWIDTH-1);
	m += HASH(n, 1);
	return m;
}

void 
dfa_init(ushort nr_layers)
{	int i; Vertex *r, *t;

	dfa_depth = nr_layers;	/* one byte per layer */
	path   = (Vertex **) emalloc((dfa_depth+1)*sizeof(Vertex *));
	layers = (Vertex **) emalloc(TWIDTH*(dfa_depth+1)*sizeof(Vertex *));
	lastword = (uchar *) emalloc((dfa_depth+1)*sizeof(uchar));
	lastword[dfa_depth] = lastword[0] = 255;
	path[0] = R = new_vertex(); F = new_vertex();

	for (i = 1, r = R; i < dfa_depth; i++, r = t)
		t = allDelta(r, i-1);
	NF = allDelta(r, i-1);
}

#if 0
static void complement_dfa(void) { Vertex *tmp = F; F = NF; NF = tmp; }
#endif

double
tree_stats(Vertex *t)
{	Edge *e; double cnt=0.0;
	if (!t) return 0;
	if (!t->key) return 0;
	t->key = 0; /* precaution */
	if (t->dst[0]) cnt++;
	if (t->dst[1]) cnt++;
	for (e = t->Succ; e; e = e->Nxt)
		cnt++;
	cnt += tree_stats(t->lnk);
	cnt += tree_stats(t->left);
	cnt += tree_stats(t->right);
	return cnt;
}

void
dfa_stats(void)
{	int i, j; double cnt = 0.0;
	for (j = 0; j < TWIDTH; j++)
	for (i = 0; i < dfa_depth+1; i++)
		cnt += tree_stats(layers[i*TWIDTH+j]);
	printf("Minimized Automaton:	%6ld nodes and %6g edges\n",
		nr_states, cnt);
}

int
dfa_member(ushort n)
{	Vertex **p, **q;
	uchar *w = &word[n];
	int i;

	p = &path[n]; q = (p+1);
	for (i = n; i < dfa_depth; i++)
		*q++ = Delta(*p++, *w++);
	return (*p == F);
}

int
dfa_store(uchar *sv)
{	Vertex **p, **q, *s, *y, *old, *new = F;
	uchar *w, *u = lastword;
	int i, j, k;

	w = word = sv;
	while (*w++ == *u++)	/* find first byte that differs */
		;
	pfrst = (int) (u - lastword) - 1;
	memcpy(&lastword[pfrst], &sv[pfrst], dfa_depth-pfrst);
	if (pfrst > iv) pfrst = iv;
	if (pfrst > nv) pfrst = nv;

	/* Phase1 */
	p = &path[pfrst]; q = (p+1); w = &word[pfrst];
	for (i = pfrst; i < dfa_depth; i++)
		*q++ = Delta(*p++, *w++);	/* (*p)->delta[*w++]; */

	if (*p == F) return 1;	/* it's already there */

	/* Phase2 */
	iv = dfa_depth;
	do {	iv--;
		old = new;
		new = find_it(path[iv], old, word[iv], iv);
	} while (new && iv > 0);

	/* Phase3 */
	nv = k = 0; s = path[0];
	for (j = 1; j <= iv; ++j) 
		if (path[j]->num > 1)
		{	y = new_vertex();
			copyEdges(y, path[j]);
			insert_it(y, j);
			numDelta(y, 1);
			delete_it(s, j-1);
			setDelta(s, word[j-1], y);
			insert_it(s, j-1);
			y->num = 1;	/* initial value 1 */
			s = y;
			path[j]->num--;	/* only 1 moved from j to y */
			k = 1;
		} else
		{	s = path[j];
			if (!k) nv = j;
		}
	y = Delta(s, word[iv]);
	y->num--;
	delete_it(s, iv); 
	setDelta(s, word[iv], old);
	insert_it(s, iv); 
	old->num++;

	for (j = iv+1; j < dfa_depth; j++)
		if (path[j]->num == 0)
		{	numDelta(path[j], -1);
			delete_it(path[j], j);
			recyc_vertex(path[j]);
		} else
			break;
	return 0;
}

static Vertex *
splay(ulong i, Vertex *t)
{	Vertex N, *l, *r, *y;

	if (!t) return t;
	N.left = N.right = (Vertex *) 0;
	l = r = &N;
	for (;;)
	{	if (i < t->key)
		{	if (!t->left) break;
			if (i < t->left->key)
			{	y = t->left;
				t->left = y->right;
				y->right = t;
				t = y;
				if (!t->left) break;
			}
			r->left = t;
			r = t;
			t = t->left;
		} else if (i > t->key)
		{	if (!t->right) break;
			if (i > t->right->key)
			{	y = t->right;
				t->right = y->left;
				y->left = t;
				t = y;
				if (!t->right) break;
			}
			l->right = t;
			l = t;
			t = t->right;
		} else
			break;
	}
	l->right = t->left;
	r->left = t->right;
	t->left = N.right;
	t->right = N.left;
	return t;
}

static void
insert_it(Vertex *v, int L)
{	Vertex *new, *t;
	ulong i; int nr;

	i = mk_key(v);
	nr = ((L*TWIDTH)+Tally);
	t = layers[nr];

	v->key = i; 
	if (!t)
	{	layers[nr] = v;
		return;
	}
	t = splay(i, t);
	if (i < t->key)
	{	new = v;
		new->left = t->left;
		new->right = t;
		t->left = (Vertex *) 0;
	} else if (i > t->key)
	{	new = v;
		new->right = t->right;
		new->left = t;
		t->right = (Vertex *) 0;
	} else	 /* it's already there */
	{	v->lnk = t->lnk; /* put in linked list off v */
		t->lnk = v;
		new = t;
	}
	layers[nr] = new;
}

static int
checkit(Vertex *h, Vertex *v, Vertex *n, uchar sigma)
{	Edge *g, *f;
	int i, k, j = 1;

	for (k = 0; k < 2; k++)
		if (h->dst[k])
		{	if (sigma >= h->from[k] && sigma <= h->to[k])
			{	if (h->dst[k] != n) goto no_match;
			}
			for (i = h->from[k]; i <= h->to[k]; i++)
			{	if (i == sigma) continue;
				g = cacheDelta(v, i, j); j = 0;
				if (h->dst[k] != g->Dst)
					goto no_match;
				if (g->s == 0 || g->S != i)
					i = g->To;
		}	}
	for (f = h->Succ; f; f = f->Nxt)
	{	if (INRANGE(f,sigma))
		{	if (f->Dst != n) goto no_match;
		}
		for (i = f->From; i <= f->To; i++)
		{	if (i == sigma) continue;
			g = cacheDelta(v, i, j); j = 0;
			if (f->Dst != g->Dst)
				goto no_match;
			if (g->s == 1 && i == g->S)
				continue;
			i = g->To;
		}
		if (f->s && f->S != sigma)
		{	g = cacheDelta(v, f->S, 1);
			if (f->Dst != g->Dst)
				goto no_match;
		}
	}
	if (h->Succ || h->dst[0] || h->dst[1]) return 1;
no_match:
	return 0;
}

static Vertex *
find_it(Vertex *v, Vertex *n, uchar sigma, int L)
{	Vertex *z, *t;
	ulong i; int nr;

	i = mk_special(sigma,n,v);
	nr = ((L*TWIDTH)+Tally);
	t = layers[nr];

	if (!t) return (Vertex *) 0;
	layers[nr] = t = splay(i, t);
	if (i == t->key)
	for (z = t; z; z = z->lnk)
		if (checkit(z, v, n, sigma))
			return z;

	return (Vertex *) 0;
}

static void
delete_it(Vertex *v, int L)
{	Vertex *x, *t;
	ulong i; int nr;

	i = cheap_key(v);
	nr = ((L*TWIDTH)+Tally);
	t = layers[nr];
	if (!t) return;

	t = splay(i, t);
	if (i == t->key)
	{	Vertex *z, *y = (Vertex *) 0;
		for (z = t; z && z != v; y = z, z = z->lnk)
			;
		if (z != v) goto bad;
		if (y)
		{	y->lnk = z->lnk;
			z->lnk = (Vertex *) 0;
			layers[nr] = t;
			return;
		} else if (z->lnk)	/* z == t == v */
		{	y = z->lnk;
			y->left = t->left;
			y->right = t->right;
			t->left = t->right = t->lnk = (Vertex *) 0;
			layers[nr] = y;
			return;
		}
		/* delete the node itself */
		if (!t->left)
		{	x = t->right;
		} else
		{	x = splay(i, t->left);
			x->right = t->right;
		}
		t->left = t->right = t->lnk = (Vertex *) 0;
		layers[nr] = x;
		return;
	}
bad:	Uerror("cannot happen delete");
}

#endif /* MA */

#if defined(W_XPT) || defined(R_XPT)

static Vertex	**temptree;
static char	wbuf[4096];
static int	WCNT = 4096, wcnt=0;
static uchar	stacker[MA+1];
static ulong	stackcnt = 0;

static void
xwrite(int fd, char *b, int n)
{
	if (wcnt+n >= 4096)
	{	write(fd, wbuf, wcnt);
		wcnt = 0;
	}
	memcpy(&wbuf[wcnt], b, n);
	wcnt += n;
}

static void
wclose(fd)
{
	if (wcnt > 0)
		write(fd, wbuf, wcnt);
	wcnt = 0;
	close(fd);
}

static void
w_vertex(int fd, Vertex *v)
{	char t[3]; int i; Edge *e;

	xwrite(fd, (char *) &v,  sizeof(Vertex *));
	t[0] = 0;
	for (i = 0; i < 2; i++)
		if (v->dst[i])
		{	t[1] = v->from[i], t[2] = v->to[i];
			xwrite(fd, t, 3);
			xwrite(fd, (char *) &(v->dst[i]), sizeof(Vertex *));
		}
	for (e = v->Succ; e; e = e->Nxt)
	{	t[1] = e->From, t[2] = e->To;
		xwrite(fd, t, 3);
		xwrite(fd, (char *) &(e->Dst), sizeof(Vertex *));

		if (e->s)
		{	t[1] = t[2] = e->S;
			xwrite(fd, t, 3);
			xwrite(fd, (char *) &(e->Dst), sizeof(Vertex *));
	}	}
}

static void
w_layer(int fd, Vertex *v)
{	uchar c=1;

	if (!v) return;
	xwrite(fd, (char *) &c, 1);
	w_vertex(fd, v);
	w_layer(fd, v->lnk);
	w_layer(fd, v->left);
	w_layer(fd, v->right);
}

void
w_xpoint(void)
{	int fd; char nm[64];
	int i, j;  uchar c;
	static uchar xwarned = 0;

	sprintf(nm, "%s.xpt", Source);
	if ((fd = creat(nm, 0666)) <= 0)
	if (!xwarned)
	{	xwarned = 1;
		printf("cannot creat checkpoint file\n");
		return;
	}
	xwrite(fd, (char *) &nstates, sizeof(double));
	xwrite(fd, (char *) &truncs, sizeof(double));
	xwrite(fd, (char *) &truncs2, sizeof(double));
	xwrite(fd, (char *) &nlinks, sizeof(double));
	xwrite(fd, (char *) &dfa_depth, sizeof(int));
	xwrite(fd, (char *) &R,  sizeof(Vertex *));
	xwrite(fd, (char *) &F,  sizeof(Vertex *));
	xwrite(fd, (char *) &NF, sizeof(Vertex *));

	for (j = 0; j < TWIDTH; j++)
	for (i = 0; i < dfa_depth+1; i++)
	{	w_layer(fd, layers[i*TWIDTH+j]);
		c = 2; xwrite(fd, (char *) &c, 1);
	}
	wclose(fd);
}

static void
xread(int fd, char *b, int n)
{	int m = wcnt; int delta = 0;
	if (m < n)
	{	if (m > 0) memcpy(b, &wbuf[WCNT-m], m);
		delta = m;
		WCNT = wcnt = read(fd, wbuf, 4096);
		if (wcnt < n-m)
			Uerror("xread failed -- insufficient data");
		n -= m;
	}
	memcpy(&b[delta], &wbuf[WCNT-wcnt], n);
	wcnt -= n;
}

static void
x_cleanup(Vertex *c)
{	Edge *e;	/* remove the tree and edges from c */
	if (!c) return;
	for (e = c->Succ; e; e = e->Nxt)
		x_cleanup(e->Dst);
	recyc_vertex(c);
}

static void
x_remove(void)
{	Vertex *tmp; int i, s;
#if 1
	int r, j;
	/* double-check: */
	stacker[dfa_depth-1] = 0; r = dfa_store(stacker);
	stacker[dfa_depth-1] = 4; j = dfa_member(dfa_depth-1);
	if (r != 1 || j != 0)
	{	printf("%ld: ", stackcnt);
		for (i = 0; i < dfa_depth; i++)
			printf("%d,", stacker[i]);
		printf(" -- not a stackstate <o:%d,4:%d>\n", r, j);
		return;
	}
#endif
	stacker[dfa_depth-1] = 1;
	s = dfa_member(dfa_depth-1);

	{ tmp = F; F = NF; NF = tmp; }	/* complement */
		if (s) dfa_store(stacker);
		stacker[dfa_depth-1] = 0;
		dfa_store(stacker);
		stackcnt++;
	{ tmp = F; F = NF; NF = tmp; }
}

static void
x_rm_stack(Vertex *t, int k)
{	int j; Edge *e;

	if (k == 0)
	{	x_remove();
		return;
	}
	if (t)
	for (e = t->Succ; e; e = e->Nxt)
	{	for (j = e->From; j <= (int) e->To; j++)
		{	stacker[k] = (uchar) j;
			x_rm_stack(e->Dst, k-1);
		}
		if (e->s)
		{	stacker[k] = e->S;
			x_rm_stack(e->Dst, k-1);
	}	}
}

static Vertex *
insert_withkey(Vertex *v, int L)
{	Vertex *new, *t = temptree[L];

	if (!t) { temptree[L] = v; return v; }
	t = splay(v->key, t);
	if (v->key < t->key)
	{	new = v;
		new->left = t->left;
		new->right = t;
		t->left = (Vertex *) 0;
	} else if (v->key > t->key)
	{	new = v;
		new->right = t->right;
		new->left = t;
		t->right = (Vertex *) 0;
	} else
	{	if (t != R && t != F && t != NF)
			Uerror("double insert, bad checkpoint data");
		else
		{	recyc_vertex(v);
			new = t;
	}	}
	temptree[L] = new;

	return new;
}

static Vertex *
find_withkey(Vertex *v, int L)
{	Vertex *t = temptree[L];
	if (t)
	{	temptree[L] = t = splay((ulong) v, t);
		if (t->key == (ulong) v)
			return t;
	}
	Uerror("not found error, bad checkpoint data");
	return (Vertex *) 0;
}

void
r_layer(int fd, int n)
{	Vertex *v;
	Edge *e;
	char c, t[2];

	for (;;)
	{	xread(fd, &c, 1);
		if (c == 2) break;
		if (c == 1)
		{	v = new_vertex();
			xread(fd, (char *) &(v->key), sizeof(Vertex *));
			v = insert_withkey(v, n);
		} else	/* c == 0 */
		{	e = new_edge((Vertex *) 0);
			xread(fd, t, 2);
			e->From = t[0];
			e->To = t[1];
			xread(fd, (char *) &(e->Dst), sizeof(Vertex *));
			insert_edge(v, e);
	}	}
}

static void
v_fix(Vertex *t, int nr)
{	int i; Edge *e;

	if (!t) return;

	for (i = 0; i < 2; i++)
	if (t->dst[i])
		t->dst[i] = find_withkey(t->dst[i], nr);

	for (e = t->Succ; e; e = e->Nxt)
		e->Dst = find_withkey(e->Dst, nr);
		
	v_fix(t->left, nr);
	v_fix(t->right, nr);
}

static void
v_insert(Vertex *t, int nr)
{	Edge *e; int i;

	if (!t) return;
	v_insert(t->left, nr);
	v_insert(t->right, nr);

	/* remove only leafs from temptree */
	t->left = t->right = t->lnk = (Vertex *) 0;
	insert_it(t, nr);	/* into layers */
	for (i = 0; i < 2; i++)
		if (t->dst[i])
			t->dst[i]->num += (t->to[i] - t->from[i] + 1);
	for (e = t->Succ; e; e = e->Nxt)
		e->Dst->num += (e->To - e->From + 1 + e->s);
}

static void
x_fixup(void)
{	int i;

	for (i = 0; i < dfa_depth; i++)
		v_fix(temptree[i], (i+1));

	for (i = dfa_depth; i >= 0; i--)
		v_insert(temptree[i], i);
}

static Vertex *
x_tail(Vertex *t, ulong want)
{	int i, yes, no; Edge *e; Vertex *v = (Vertex *) 0;

	if (!t) return v;

	yes = no = 0;
	for (i = 0; i < 2; i++)
		if ((ulong) t->dst[i] == want)
		{	if (t->from[i] <= 0 && t->to[i] >= 0)
				yes = 1;
			else
			if (t->from[i] <= 4 && t->to[i] >= 4)
				no = 1;
		}

	for (e = t->Succ; e; e = e->Nxt)
		if ((ulong) e->Dst == want)
		{	if (INRANGE(e, 0))
				yes = 1;
			else if (INRANGE(e, 4))
				no = 1;
		}
	if (yes && !no) return t;
	v = x_tail(t->left, want);  if (v) return v;
	v = x_tail(t->right, want); if (v) return v;
	return (Vertex *) 0;
}

static void
x_anytail(Vertex *t, Vertex *c, int nr)
{	int i; Edge *e, *f; Vertex *v;

	if (!t) return;

	for (i = 0; i < 2; i++)
		if ((ulong) t->dst[i] == c->key)
		{	v = new_vertex(); v->key = t->key;
			f = new_edge(v);
			f->From = t->from[i];
			f->To = t->to[i];
			f->Nxt = c->Succ;
			c->Succ = f;
			if (nr > 0)
			x_anytail(temptree[nr-1], v, nr-1);
		}

	for (e = t->Succ; e; e = e->Nxt)
		if ((ulong) e->Dst == c->key)
		{	v = new_vertex(); v->key = t->key;
			f = new_edge(v);
			f->From = e->From;
			f->To = e->To;
			f->s = e->s;
			f->S = e->S;
			f->Nxt = c->Succ;
			c->Succ = f;
			x_anytail(temptree[nr-1], v, nr-1);
		}

	x_anytail(t->left, c, nr);
	x_anytail(t->right, c, nr);
}

static Vertex *
x_cpy_rev(void)
{	Vertex *c, *v;

	/* find 0 and !4 predecessor of F */

	v = x_tail(temptree[dfa_depth-1], F->key);
	if (!v) return (Vertex *) 0;

	c = new_vertex(); c->key = v->key;

	/* every node on dfa_depth-2 that has v->key as succ */
	/* make copy and let c point to these (reversing ptrs) */

	x_anytail(temptree[dfa_depth-2], c, dfa_depth-2);
 
	return c;
}

void
r_xpoint(void)
{	int fd; char nm[64]; Vertex *d;
	int i, j;

	wcnt = 0;
	sprintf(nm, "%s.xpt", Source);
	if ((fd = open(nm, 0)) < 0)	/* O_RDONLY */
		Uerror("cannot open checkpoint file");

	xread(fd, (char *) &nstates,   sizeof(double));
	xread(fd, (char *) &truncs,    sizeof(double));
	xread(fd, (char *) &truncs2,   sizeof(double));
	xread(fd, (char *) &nlinks,    sizeof(double));
	xread(fd, (char *) &dfa_depth, sizeof(int));

	if (dfa_depth != MA+a_cycles)
		Uerror("bad dfa_depth in checkpoint file");

	path	  = (Vertex **) emalloc((dfa_depth+1)*sizeof(Vertex *));
	layers	  = (Vertex **) emalloc(TWIDTH*(dfa_depth+1)*sizeof(Vertex *));
	temptree  = (Vertex **) emalloc((dfa_depth+2)*sizeof(Vertex *));
	lastword  = (uchar *)   emalloc((dfa_depth+1)*sizeof(uchar));
	lastword[dfa_depth] = lastword[0] = 255; 

	path[0] = R = new_vertex();
	xread(fd, (char *) &R->key, sizeof(Vertex *));
	R = insert_withkey(R, 0);

	F = new_vertex();
	xread(fd, (char *) &F->key, sizeof(Vertex *));
	F = insert_withkey(F, dfa_depth);

	NF = new_vertex();
	xread(fd, (char *) &NF->key, sizeof(Vertex *));
	NF = insert_withkey(NF, dfa_depth);

	for (j = 0; j < TWIDTH; j++)
	for (i = 0; i < dfa_depth+1; i++)
		r_layer(fd, i);

	if (wcnt != 0) Uerror("bad count in checkpoint file");

	d = x_cpy_rev();
	x_fixup();
	stacker[dfa_depth-1] = 0;
	x_rm_stack(d, dfa_depth-2);
	x_cleanup(d);
	close(fd);

	printf("pan: removed %ld stackstates\n", stackcnt);
	nstates -= (double) stackcnt;
}

#endif /* defined(W_XPT) || defined(R_XPT) */


