/***** dataflow.c  --  Data flow analysis routines *****/

/* Data flow engine written by Radu Iosif <iosif@athena.polito.it>         */
/* This code is based on algorithms described in the "Dragon Book"         */
/* by Aho, Sethi and Ullman. Data flow problem is to compute an alias      */
/* list for every used variable in the program. This information is        */
/* needed in order to assume that a transition is "safe".                  */

/* dSpin extension written by Radu Iosif. Feedback: iosif@athena.polito.it */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include "dfa.h"
#ifdef PC
#include "y_tab.h"
#else
#include "y.tab.h"
#endif

extern int no_dfa;

#define INC(ptr, type, no)                                 \
do {                                                       \
  (no) ++;                                                 \
  (ptr) = ((type*) erealloc ((ptr), (no) * sizeof(type))); \
} while (0)


#define DEC(ptr, type, no)                                 \
do {                                                       \
  (no) --;                                                 \
  (ptr) = ((type*) erealloc ((ptr), (no) * sizeof(type))); \
} while (0)


/* Some descriptor tables */

static int     s_idx = 0;       /* symbol */

static int     v_idx = 0;       /* variable */
static var_t  *v_tab = NULL;

static int     f_idx = 0;       /* defun */
static fun_t  *f_tab = NULL;


#define HEAP             v_idx + f_idx

#define V_TAB(x)         v_tab [(x) - f_idx]
#define F_TAB(x)         f_tab [(x)]

#define V_MIN            f_idx
#define V_MAX            f_idx + v_idx

#define F_MIN            0
#define F_MAX            f_idx


/* The idx of the function we are currently partitioning */
/* Needed by lextok_idx() and trans_el()                 */

static int     Fdx;


/*** Basic block management routines ***/

static bl_t bl_create()
{
  return (bl_t) emalloc (sizeof(struct bb_list));
}


static void bl_delete(bl_t bl)
{
  free(bl);
}


static void bb_add(bb_t block, Element *el)
{

#if DEBUG>=2
  printf("BB_add [");
  comment(stdout, el->n, 1);
  printf("] to (%d)\n", block->bid);
#endif /* DEBUG */

  INC(block->e_seq, Element *, block->e_no);  
  block->e_seq [block->e_no - 1] = el;

  el->block = block;
}


static void bb_edge(bb_t from, bb_t to)
{
  INC(from->succ_list, bb_t, from->succ_no);
  from->succ_list [from->succ_no - 1] = to;

  INC(to->pred_list, bb_t, to->pred_no);
  to->pred_list [to->pred_no - 1] = from;
}


static bb_t bb_create(bl_t bl, Element *leader)
{
  int i;

  for (i = 0; i < NO(bl); i ++)
    {
      if (LS(bl) [i]->e_seq [0] == leader)
	return LS(bl) [i];
    }

  INC(LS(bl), bb_t, NO(bl));

  LS(bl) [NO(bl) - 1] = (bb_t) emalloc (sizeof(struct bblock));
  LS(bl) [NO(bl) - 1]->bid = NO(bl);

#if DEBUG>=2
  printf("BB_leader of (%d) --> [", NO(bl));
  comment(stdout, leader->n, 1);
  printf("]\n");
#endif /* DEBUG */

  bb_add(LS(bl) [NO(bl) - 1], leader);
  return LS(bl) [NO(bl) - 1];
}


static void bb_reach(bl_t bl, char *name)
{
  int i, j, warned = 0;

  for (i = 0; i < NO(bl); i ++)
    {
      bb_t block = LS(bl) [i];
      
      if (!block->mark)
	{
	  for (j = 0; j < block->e_no; j ++)
	    {
	      Element *e = block->e_seq [j];

	      switch (e->n->ntyp)
		{
		case BREAK:         case '.': case ATOMIC: 
		case NON_ATOMIC: case D_STEP:  case IF: case DO:
		  continue;
		}

	      if (! warned ++)
		printf("unreachable in %s():\n", name);

	      if (e->n->fn)
		printf("   %s %d: ", e->n->fn->name, e->n->ln);
	      else
		printf("  ");

	      comment(stdout, e->n, 1);
	      printf("\n");
	    }
	}
    }
}


#if DEBUG>=2

static void a_printf(ss_t);
static void l_printf(ss_t);

static void bb_printf(bl_t bl)
{
  int i, j;

  for (i = NO(bl) - 1; i >= 0; i --)
    {
      bb_t block = LS(bl) [i];

      printf("BB(%d):\n", block->bid);

      printf("^[");
      for (j = 0; j < block->pred_no; j ++)
	printf("%d ", block->pred_list [j]->bid);
      printf("]\n");

      printf("v[");
      for (j = 0; j < block->succ_no; j ++)
	printf("%d ", block->succ_list [j]->bid);
      printf("]\n");

      for (j = 0; j < block->e_no; j ++)
	{
	  printf("\t"); 
	  comment(stdout, block->e_seq [j]->n, 1);
	  printf("\n");
	}

      printf("*** IN ***\n");
      a_printf(block->in);
      printf("\n");

      printf("*** OUT ***\n");
      a_printf(block->out);
      printf("\n");

      printf("--- IN_LN ---\n");
      l_printf(block->in_ln);
      printf("\n");

      printf("--- OUT_LN ---\n");
      l_printf(block->out_ln);
      printf("\n");
    }
}

#endif /* DEBUG */


/*** Basic block partitioning routines ***/

static void lextok_idx(Lextok *);
static void init_idx(Sequence *);

static void bb_leader(bl_t bl, Element *el, int start)
{
  if (!el)
    return;

  if (el->vis == 1)
    return;
  else
    el->vis = 1;

  if (start || el->esc)
    bb_create(bl, el);	

  /* hooked this stuff here :-) */

  if (el->tag & BB)
    {
      enter_seq(el->sbl->this);
      init_idx(el->sbl->this);
    }

  lextok_idx(el->n);

  if (el->tag & BE)
    exit_seq();

  switch (el->n->ntyp)
    {
    case UNLESS:
      {
	SeqList *sl = el->n->sl;
	Element *e;

	enter_seq(sl->this);
	bb_leader(bl, sl->this->frst, 0);
	exit_seq();

	bb_leader(bl, sl->nxt->this->frst, 1);
	break;
      }

    case ATOMIC:
    case D_STEP:
    case NON_ATOMIC:

      last_el(el->n->sl->this)->nxt = el->nxt;
      bb_leader(bl, el->n->sl->this->frst, 0);

      break;

    case IF:
      {
	SeqList *sl;

	for (sl = el->n->sl; sl; sl = sl->nxt)
	  bb_leader(bl, sl->this->frst, 1);

	break;
      }

    case DO:
      {
	SeqList *sl;

	for (sl = el->n->sl; sl; sl = sl->nxt)
	  bb_leader(bl, sl->this->frst, 1);

	bb_leader(bl, el->nxt, 1);
	break;
      }

    case '.':

      if (!start)
	bb_create(bl, el);

      bb_leader(bl, el->nxt, 0);

      break;

    case GOTO:

      bb_create(bl, target(el));
      bb_leader(bl, el->nxt, 1);

      break;

    case RETURN:
    case RETFUN:

      bb_leader(bl, el->nxt, 1);
      break;

    default:

      bb_leader(bl, el->nxt, 0);
    }
}


static void bb_setup(bb_t block, Element *el)
{
  if (!el)
    return;

  if (el->vis == 2)
    return;
  else
    el->vis = 2;

  if (!el->block)
    bb_add(block, el);
  else
    block = el->block;

  switch (el->n->ntyp)
    {
    case UNLESS:

      bb_setup(NULL, el->n->sl->this->frst);
      bb_setup(NULL, el->n->sl->nxt->this->frst);

      break;

    case ATOMIC:
    case D_STEP:
    case NON_ATOMIC:
      
      bb_setup(block, el->n->sl->this->frst);
      bb_setup(last_el(el->n->sl->this)->block, el->nxt);

      break;

    case IF:
    case DO:
      {
	SeqList *sl;

	for (sl = el->n->sl; sl; sl = sl->nxt)
	  bb_setup(block, sl->this->frst);

	bb_setup(block, el->nxt);

	break;
      }

    default:

      bb_setup(block, el->nxt);
    }
}


static void bb_link(bl_t bl)
{
  Element *last;
  SeqList *esc;
  int i;

  for (i = 0; i < NO(bl); i ++)
    {
      bb_t block = LS(bl) [i];

      last  = block->e_seq [block->e_no - 1];

      for (esc = last->esc; esc; esc = esc->nxt)
	bb_edge(block, frst_el(esc->this)->block);

      switch (last->n->ntyp)
	{
	case UNLESS:
	  {
	    SeqList *sl = last->n->sl;

	    bb_edge(block, frst_el(sl->this)->block);
	    bb_edge(block, frst_el(sl->nxt->this)->block);

	    break;
	  }
				   
	case IF:
	case DO:
	  {
	    SeqList *sl;

	    for (sl = last->n->sl; sl; sl = sl->nxt) 
	      bb_edge(block, frst_el(sl->this)->block);

	    break;
	  }
	  
	case GOTO:
	  bb_edge(block, target(last)->block);
	  break;

	  /* 
	   * An exception occurs when we 
	   * use one of those inside unless.
	   */
	  
	case ATOMIC:
	case NON_ATOMIC:
	case D_STEP:
	  {
	    Element *frst = frst_el(last->n->sl->this);

	    if (block != frst->block)
	      bb_edge(block, frst->block);
	  }

	case RETURN:
	case RETFUN:

	  break;

	default:

	  if (last->nxt)
	    bb_edge(block, last->nxt->block);
	}
    }
}


/*** Depth first ordering routine ***/

bb_order(bb_t block, bl_t bl)
{
  bb_t succ;
  int i;

  block->mark = 1;

  for (i = 0; i < block->succ_no; i ++)
    {
      succ = block->succ_list [i];

      if (!succ->mark)
	bb_order(succ, bl);
    }

  INC(LS(bl), bb_t, NO(bl));
  LS(bl) [NO(bl) - 1] = block;
}


/*** Set management routines ***/

static set_t set_create(int no)
{
  set_t set = (set_t) emalloc (sizeof(struct set));
  int size;

  if (no % (WSIZE * 8))
    size = no / (WSIZE * 8) + 1;
  else
    size = no / (WSIZE * 8);

  set->max  = size * WSIZE * 8;
  set->size = size * WSIZE;
  set->bmap = (WTYPE *) emalloc (size * WSIZE);

  return set;
}


static void set_delete(set_t set)
{
  if (set->bmap)
    free(set->bmap);

  free(set);
}


static void set_clear(set_t set)
{
  int i;

  for (i = 0; i < set->size; i ++)
    set->bmap [i] = 0;
}


static void set_copy(set_t dst, set_t src)
{
  int i;

  assert(dst->size == src->size);

  for (i = 0; i < dst->size; i ++)
    dst->bmap [i] = src->bmap [i];
}


static void set_union(set_t dst, set_t src)
{
  int i;

  assert(dst->size == src->size);

  for (i = 0; i < dst->size; i ++)
    dst->bmap [i] |= src->bmap [i];
}


static void set_add(set_t set, int pos)
{
  assert(pos < set->max);

  set->bmap [pos / (WSIZE * 8)] |= 1 << (pos % (WSIZE * 8));
}


static void set_del(set_t set, int pos)
{
  assert(pos < set->max);

  set->bmap [pos / (WSIZE * 8)] &= ~(1 << (pos % (WSIZE * 8)));
}


static int set_mem(set_t set, int pos)
{
  assert(pos < set->max);

  return set->bmap [pos / (WSIZE * 8)] & (1 << (pos % (WSIZE * 8)));
}


static int set_sub(set_t sub, set_t set)
{
  int i;

  assert(set->size == sub->size);

  for (i = 0; i < set->size; i ++)
    {
      if ((set->bmap [i] | sub->bmap [i]) != set->bmap [i])
	return 0;
    }

  return 1;
}


static set_t set_diff(set_t one, set_t two)
{
  set_t res;
  int i;

  assert(one->size == two->size);

  res = set_create(one->max);
  
  for (i = 0; i < one->size; i ++)
    res->bmap [i] = one->bmap [i] & ~two->bmap [i];
  
  return res;
}


static int set_equal(set_t one, set_t two)
{
  int i;

  if (one->size != two->size)
    return 0;

  for (i = 0; i < one->size; i ++)
    {
      if (one->bmap [i] != two->bmap [i])
	return 0;
    }

  return 1;
}


#if DEBUG>=1

static void set_printf(set_t set)
{
  WTYPE mask;
  int i;

  printf("[");

  for (i = 0; i < set->max; i ++)
    {
      mask = 1 << (i % WSIZE);
      printf("%d", (set->bmap [i / (WSIZE * 8)] & mask) >> (i % (WSIZE * 8)));
    }

  printf("]");
}

#endif /* DEBUG */


/*** Superset management routines ***/

static ss_t ss_create(int no, int sz)
{
  ss_t sset = (ss_t) emalloc (sizeof(struct superset));
  int i;

  sset->s_no  = no;
  sset->s_set = (set_t *) emalloc (no * sizeof(set_t));

  for (i = 0; i < no; i ++)
    sset->s_set [i] = set_create(sz);
  
  return sset;
}


static ss_t a_create()
{
  return ss_create(v_idx + f_idx + 1, v_idx + f_idx + 1);
}


static ss_t l_create()
{
  return ss_create(v_idx + f_idx, v_idx + f_idx);
}


static void ss_delete(ss_t sset)
{
  int i;

  for (i = 0; i < sset->s_no; i ++)
    set_delete(sset->s_set [i]);

  if (sset->s_set)
    free(sset->s_set);

  free(sset);
}


static void ss_copy(ss_t dst, ss_t src)
{
  int i;

  assert(dst->s_no == src->s_no);

  for (i = 0; i < dst->s_no; i ++)    
    set_copy(dst->s_set [i], src->s_set [i]);
}


static int ss_equal(ss_t one, ss_t two)
{
  int i;

  assert(one->s_no == two->s_no);

  for (i = 0; i < one->s_no; i ++)
    {
      if (!set_equal(one->s_set [i], two->s_set [i]))
	return 0;
    }

  return 1;
}


static void ss_merge(ss_t dst, ss_t src)
{
  int i;

  assert(dst->s_no == src->s_no);

  for (i = 0; i < dst->s_no; i ++)
    {
      set_union(dst->s_set [i], src->s_set [i]);
    }
}


#if DEBUG>=1

static void a_printf(ss_t sset)
{
  int i, j;

  for (i = V_MIN; i < V_MAX; i ++)
    {
      comment(stdout, V_TAB(i)->tok, 1);
      printf(": ");

      for (j = V_MIN; j < V_MAX; j ++)
	{
	  if (set_mem(sset->s_set [i], j))
	    {
	      comment(stdout, V_TAB(j)->tok, 1);
	      printf(" ");
	    }	  
	}

      for (j = F_MIN; j < F_MAX; j ++)
	{
	  if (set_mem(sset->s_set [i], j))
	    printf("%s() ", F_TAB(j)->fun->fun->name);
	}

      if (set_mem(sset->s_set [i], HEAP))
	printf("[heap]");

      printf("\n");
    }
}


static void l_printf(ss_t ssln)
{
  int i, j;

  for (i = V_MIN; i < V_MAX; i ++)
    {
      comment(stdout, V_TAB(i)->tok, 1);
      printf(": ");

      for (j = V_MIN; j < V_MAX; j ++)
	{
	  if (set_mem(ssln->s_set [i], j))
	    {
	      comment(stdout, V_TAB(j)->tok, 1);
	      printf(" ");
	    }	  
	}

      printf("\n");
    }
}

#endif /* DEBUG */


/*** Link management routines ***/

static ln_t ln_create()
{
  ln_t ln;

  ln = (ln_t) emalloc (sizeof(struct link));

  return ln;
}


static void ln_delete(ln_t ln)
{
  int i;

  for (i = 0; i < ln->l_no; i ++)
    set_delete(ln->l_list [i].set);

  if (ln->l_list)
    free(ln->l_list);

  free(ln);
}


static void ln_add(ln_t ln, int idx, set_t set)
{
  INC(ln->l_list, struct ln_tag, ln->l_no);
  
  ln->l_list [ln->l_no - 1].idx = idx;
  ln->l_list [ln->l_no - 1].set = set;
}


/*** Variable indexing routines ***/

static int set_idx(set_t set)
{
  int i;

  for (i = 0; i < v_idx; i ++)
    {
      if (set_equal(v_tab [i]->set, set))
	return i + f_idx;
    }

  return -1;
}


static int var_idx(set_t var, Lextok *n, int glob, int ptr, int fptr)
{
  int i;

  for (i = 0; i < v_idx; i ++)
    {
      if (set_equal(v_tab [i]->set, var))
	return i + f_idx;
    }

  INC(v_tab, var_t, v_idx);

  v_tab [i] = (var_t) emalloc (sizeof(struct var_tag));

  v_tab [i]->set  = var;
  v_tab [i]->tok  = n;
  v_tab [i]->glob = glob;
  v_tab [i]->ptr  = ptr;
  v_tab [i]->fptr = fptr;

  return i + f_idx;
}


extern Symbol *do_same2(Lextok *, Symbol *);

static void name_set(Lextok *n, Symbol *s, set_t t)
{
  Symbol *v;

  if (!s)
    return;

  set_add(t, s->idx);

  if (!n->rgt 
      || !n->rgt->lft)
    return;

  if (!(v = do_same2(n, s)))
    return;

  name_set(n->rgt->lft, v, t);
}


static Symbol * name_last(Lextok *n, Symbol *s)
{
  if (!s) {
    printf("%s %d: cannot happen (name_last)\n", __FILE__, __LINE__);
    exit(1);
  }

  if (!n->rgt
      || !n->rgt->lft)
    return s;

  return name_last(n->rgt->lft, do_same2(n, s));
}


static void name_idx(Lextok *n)
{
  Symbol *s, *v;
  set_t set = set_create(s_idx);

  if (!n->sym)
    fatal("no name - struct_idx", 0);

  if (strcmp(n->sym->name, "_pid") == 0
      || strcmp(n->sym->name, "_p") == 0
      || strcmp(n->sym->name, "_") == 0)
    return;

  if (!(s = findloc2(n->sym)))
    s = n->sym;

  name_set(n, s, set);

  v = name_last(n, s);

  n->idx = var_idx(set, n, 
		   s->context ? 0 : 1,         /* global  */ 
		   v->type & POINTER ? 1 : 0,  /* pointer */
		   v->type == FTYPE ? 1 : 0);
  
#if DEBUG>=2
  comment(stdout, n, 1);
  printf(" --> %d\n", n->idx);
#endif /* DEBUG */  
}


static void fname_idx(Lextok *n)
{
  n->idx = fun_idx(getdefun(n->sym));
}


static void lextok_idx(Lextok *n)
{
  if (!n)
    return;

  switch (n->ntyp)
    {
    case NAME:
      
      name_idx(n);
      break;

    case FNAME:

      fname_idx(n);
      break;

    case FUNCALL:
      {
	Lextok *v;
	
	lextok_idx(n->lft);
	
	lextok_idx(n->rgt->lft);
	lextok_idx(n->rgt->rgt);

	for (v = n->rgt->lft; v; v = v->rgt)
	  lextok_idx(v->lft);
	
	break;
      }

    case RETURN:

      n->idx = Fdx;
      lextok_idx(n->lft);

      break;

    case RETFUN:
      {
	Lextok *v;

	lextok_idx(n->rgt->lft);
	lextok_idx(n->rgt->rgt);
	
	for (v = n->rgt->lft; v; v = v->rgt)
	  lextok_idx(v->lft);

	n->idx = Fdx;	

	break;
      }

    case 's': case 'r': case 'R':
      {
	Lextok *v;

	lextok_idx(n->lft);
      
	for (v = n->rgt; v; v = v->rgt)
	  {
	    if (v->lft->ntyp == CONST)
	      continue;

	    if (v->lft->ntyp == EVAL)
	      lextok_idx(v->lft->lft);
	    else
	      lextok_idx(v->lft);
	  }

	break;
      }

    case NEMPTY: case NFULL: case FULL: case EMPTY: case LEN:
    case LVAL: case '!' : case UMIN: case '~': case ASSERT:
    case NEW: case DELETE:

      lextok_idx(n->lft);
      break;

    case '/': case '*': case '-': case '+':
    case '%': case LT:  case GT: case '&': case '^':
    case '|': case LE:  case GE:  case NE: case '?':
    case EQ:  case OR:  case AND: case LSHIFT:
    case RSHIFT: case 'c': case ASGN:

      lextok_idx(n->lft);
      lextok_idx(n->rgt);

      break;

    case PRINT:
      {
	Lextok *v;

	for (v = n->lft; v; v = v->rgt)
	  lextok_idx(v->lft);
      }
    }
}


static void formal_idx(Lextok *prm)
{
  Lextok *lst, *par;

  for (lst = prm; lst; lst = lst->rgt)
    for (par = lst->lft; par; par = par->rgt)
      {
	lextok_idx(par);
      }
}


static void init_idx(Sequence *s)
{
  Symbol *sp;

  for (sp = s->symtab; sp; sp = sp->next)
    {
      if (sp->type == LABEL)
	continue;

      if (!(sp->hidden & 1)
	  && !(sp->hidden & 16))
	{	  
	  lextok_idx(sp->ini);
	}
    }
}


#if DEBUG>=1

static void ix_printf()
{
  int i, j;

  for (i = V_MIN; i < V_MAX; i ++)
    {
      printf("(%d) ", i);
      comment(stdout, V_TAB(i)->tok, 1);
      printf("\t");

      set_printf(V_TAB(i)->set);

      if (V_TAB(i)->glob)
	printf("\tglobal");

      if (V_TAB(i)->ptr)
	printf("\tpointer");

      printf("\n");
    }

  printf("\n");
}

#endif /* DEBUG */


/*** Alias management routines ***/

static void alias_kill(ss_t ss_lft, ss_t ss_rgt,
		       ss_t ssln, int lft, int rgt)
{
  int idx, lnk;

  assert(ss_lft->s_no == ss_rgt->s_no);

  for (idx = 0; idx < ss_lft->s_no; idx ++)
    set_del(ss_lft->s_set [idx], lft);

  set_clear(ss_lft->s_set [lft]);
  
  set_add(ss_lft->s_set [lft], rgt);
  set_add(ss_rgt->s_set [rgt], lft);

  /* and now the same for all links of lft */

  for (lnk = 0; lnk < ssln->s_no; lnk ++)
    {
      if (set_mem(ssln->s_set [lft], lnk))
	{
	  for (idx = 0; idx < ss_lft->s_no; idx ++)
	    set_del(ss_lft->s_set [idx], lnk);

	  set_clear(ss_lft->s_set [lnk]);

	  set_add(ss_lft->s_set [lnk], rgt);
	  set_add(ss_rgt->s_set [rgt], lnk);
	}
    }
}


static void alias_nokill(ss_t ss_lft, ss_t ss_rgt, 
			 ss_t ssln, int lft, int rgt)
{
  int lnk;

  assert(ss_lft->s_no == ss_rgt->s_no);

  set_add(ss_lft->s_set [lft], rgt);
  set_add(ss_rgt->s_set [rgt], lft);
  
  /* and now the same for all links of lft */

  for (lnk = 0; lnk < ssln->s_no; lnk ++)
    {
      if (set_mem(ssln->s_set [lft], lnk))
	{
	  set_add(ss_lft->s_set [lnk], rgt);
	  set_add(ss_rgt->s_set [rgt], lnk);
	}
    }
}


static void replace_kill(ss_t ss_lft, ss_t ss_rgt,
			 ss_t ssln, int lft, int rgt)
{
  int idx, lnk;

  assert(ss_lft->s_no == ss_rgt->s_no);

  for (idx = 0; idx < ss_rgt->s_no; idx ++)
    {
      set_del(ss_lft->s_set [idx], lft);

      if (set_mem(ss_rgt->s_set [rgt], idx))
	set_add(ss_lft->s_set [idx], lft);
    }

  set_copy(ss_lft->s_set [lft], ss_rgt->s_set [rgt]);

  /* and now the same for all links of lft */

  for (lnk = 0; lnk < ssln->s_no; lnk ++)
    {
      if (set_mem(ssln->s_set [lft], lnk))
	{
	  for (idx = 0; idx < ss_rgt->s_no; idx ++)
	    {
	      set_del(ss_lft->s_set [idx], lnk);

	      if (set_mem(ss_rgt->s_set [rgt], idx))
		set_add(ss_lft->s_set [idx], lnk);
	    }
	}
    }
}


static void replace_nokill(ss_t ss_lft, ss_t ss_rgt, 
			   ss_t ssln, int lft, int rgt)
{
  int idx, lnk;

  assert(ss_lft->s_no == ss_rgt->s_no);

  for (idx = 0; idx < ss_rgt->s_no; idx ++)
    {
      if (set_mem(ss_rgt->s_set [rgt], idx))
	set_add(ss_lft->s_set [idx], lft);
    }

  set_union(ss_lft->s_set [lft], ss_rgt->s_set [rgt]);

  /* and now the same for all links of lft */

  for (lnk = 0; lnk < ssln->s_no; lnk ++)
    {
      if (set_mem(ssln->s_set [lft], lnk))
	{
	  for (idx = 0; idx < ss_lft->s_no; idx ++)
	    {
	      if (set_mem(ss_rgt->s_set [rgt], idx))
		set_add(ss_lft->s_set [idx], lnk);
	    }

	  set_union(ss_lft->s_set [lnk], ss_rgt->s_set [rgt]);
	}
    }
}


static ln_t lnk_of(int ix)
{
  ln_t ln = ln_create();
  int i;

  for (i = V_MIN; i < V_MAX; i ++)
    {
      if (i == ix)
	continue;
      
      if (set_sub(V_TAB(ix)->set, V_TAB(i)->set))
	ln_add(ln, i, set_diff(V_TAB(i)->set, V_TAB(ix)->set));
    }

  return ln;
}


static void lnk_do_kill(ss_t ssln, int lft, int rgt)
{
  ln_t lft_ln = lnk_of(lft);
  ln_t rgt_ln = lnk_of(rgt);
  int i, j;

  for (i = 0; i < lft_ln->l_no; i ++)
    for (j = 0; j < rgt_ln->l_no; j ++)
      {
	if (set_equal(lft_ln->l_list [i].set, rgt_ln->l_list [j].set))
	  {
	    int lx = lft_ln->l_list [i].idx;
	    int rx = rgt_ln->l_list [j].idx;
	    int idx;
	    
	    for (idx = 0; idx < v_idx; idx ++)
	      {
		set_del(ssln->s_set [idx], lx);
	      }

	    set_clear(ssln->s_set [lx]);

	    set_add(ssln->s_set [lx], rx);
	    set_add(ssln->s_set [rx], lx);
	  }
      }

  ln_delete(lft_ln);
  ln_delete(rgt_ln);
}


static void lnk_do_nokill(ss_t ssln, int lft, int rgt)
{
  ln_t lft_ln = lnk_of(lft);
  ln_t rgt_ln = lnk_of(rgt);
  int i, j;

  for (i = 0; i < lft_ln->l_no; i ++)
    for (j = 0; j < rgt_ln->l_no; j ++)
      {
	if (set_equal(lft_ln->l_list [i].set, rgt_ln->l_list [j].set))
	  {
	    int lx = lft_ln->l_list [i].idx;
	    int rx = rgt_ln->l_list [j].idx;

	    set_add(ssln->s_set [lx], rx);
	    set_add(ssln->s_set [rx], lx);
	  }
      }

  ln_delete(lft_ln);
  ln_delete(rgt_ln);
}


static void call_merge(Lextok *args, int fx, ss_t sset, ss_t ssln)
{
  Lextok *act, *par, *lst;

  Defun *fn = F_TAB(fx)->fun;

  ss_t in    = F_TAB(fx)->in;
  ss_t in_ln = F_TAB(fx)->in_ln;


  for (lst = fn->prms, act = args; lst; lst = lst->rgt)
    for (par = lst->lft; par && act; par = par->rgt, act = act->rgt)
      {       
	int lx = par->idx;
	int rx;

	if (V_TAB(lx)->ptr)
	  {
	    if (act->lft->ntyp == LVAL)
	      {
		rx = act->lft->lft->idx;

		alias_nokill(in, sset, ssln, lx, rx);
		lnk_do_nokill(in_ln, lx, rx);
	      }
	    else
	      if (act->lft->ntyp == NAME)
		{
		  rx = act->lft->idx;

		  if (V_TAB(rx)->ptr)
		    {
		      replace_nokill(in, sset, ssln, lx, rx);
		      lnk_do_nokill(in_ln, lx, rx);
		    }
		}
	  }
	else
	  if (V_TAB(lx)->fptr)
	    {
	      rx = act->lft->idx;

	      if (rx >= V_MIN && V_TAB(rx)->fptr)
		replace_nokill(in, sset, ssln, lx, rx);
	      else
		alias_nokill(in, sset, ssln, lx, rx);
	    }
      }
}


static void retn_merge(Lextok *ret, int fx, ss_t sset, ss_t ssln)
{
  int idx, lx, rx;
  
  ss_t out   = f_tab [fx]->out;
  set_t rset = f_tab [fx]->rset;

  if (!ret)
    return;

  lx = ret->idx;

  /* unless array kill all previous definitions */

  if (!ret->lft)
    {
      for (idx = 0; idx < sset->s_no; idx ++)
	set_del(sset->s_set [idx], lx);

      set_clear(sset->s_set [lx]);
    }

  /* iterate over all possible returns */

  for (rx = 0; rx < rset->max; rx ++)
    {
      if (set_mem(rset, rx))
	{
	  if (V_TAB(lx)->ptr)
	    {
	      if (V_TAB(rx)->ptr)
		{
		  replace_nokill(sset, out, ssln, lx, rx);
		  lnk_do_nokill(ssln, lx, rx);
		}
	      else
		alias_nokill(sset, out, ssln, lx, rx);	      
	    }
	  else
	    if (V_TAB(lx)->fptr)
	      {
		if (rx >= V_MIN && V_TAB(rx)->fptr)
		  replace_nokill(sset, out, ssln, lx, rx);
		else
		  alias_nokill(sset, out, ssln, lx, rx);
	      }
	}
    }
}


static void scope_merge(Sequence *s, ss_t sset, ss_t ssln)
{
  set_t set = set_create(s_idx);
  Symbol *sp;
  int lx, rx;

  for (sp = s->symtab; sp; sp = sp->next)
    {
      if (!sp->ini)
	continue;

      set_clear(set);
      set_add(set, sp->idx);

      if ((lx = set_idx(set)) < 0)
	continue;      

      switch (sp->ini->ntyp)
	{
	case LVAL:

	  rx = sp->ini->lft->idx;

	  if (V_TAB(lx)->ptr)
	    {
	      assert(lx != rx);

	      alias_nokill(sset, sset, ssln, lx, rx);
	      lnk_do_nokill(ssln, lx, rx);
	    }

	  break;

	case NAME:

	  rx = sp->ini->idx;

	  if (lx == rx)
	    break;

	  if (V_TAB(lx)->ptr 
	      && V_TAB(rx)->ptr)
	    {
	      replace_nokill(sset, sset, ssln, lx, rx);
	      lnk_do_nokill(ssln, lx, rx);
	    }
	  else
	    if (V_TAB(lx)->fptr 
		&& V_TAB(rx)->fptr)
	      {
		replace_nokill(sset, sset, ssln, lx, rx);
	      }

	  break;

	case FNAME:

	  rx = sp->ini->idx;

	  if (V_TAB(lx)->fptr)
	    {
	      alias_nokill(sset, sset, ssln, lx, rx);
	    }
	}
    }
}


static void trans_el(Element *el, ss_t sset, ss_t ssln)
{
  int rx, lx, fx;
  int array;

  if (el->tag & BB)
    scope_merge(el->sbl->this, sset, ssln);
  
  switch (el->n->ntyp)
    {
    case ASGN:
      {
	array = el->n->lft->lft ? 1 : 0;
	lx    = el->n->lft->idx;
	
	if (V_TAB(lx)->ptr)
	  {
	    switch (el->n->rgt->ntyp)
	      {
	      case LVAL:
	    
		rx = el->n->rgt->lft->idx;

		assert(lx != rx);

		if (array)
		  {
		    alias_nokill(sset, sset, ssln, lx, rx);
		    lnk_do_nokill(ssln, lx, rx);
		  }
		else
		  {
		    alias_kill(sset, sset, ssln, lx, rx);
		    lnk_do_kill(ssln, lx, rx);
		  }

		break;
	    
	      case NAME:
		
		rx = el->n->rgt->idx;	    
		
		if (!V_TAB(rx)->ptr
		    || lx == rx)
		  break;

		if (array)
		  {
		    replace_nokill(sset, sset, ssln, lx, rx);
		    lnk_do_nokill(ssln, lx, rx);
		  }
		else
		  {
		    replace_kill(sset, sset, ssln, lx, rx);
		    lnk_do_kill(ssln, lx, rx);
		  }
	      }
	  }      /* assignment to pointer */
	else
	  if (V_TAB(lx)->fptr)
	    {
	      rx = el->n->rgt->idx;

	      switch (el->n->rgt->ntyp)
		{
		case NAME:
		  
		  if (!V_TAB(rx)->fptr
		      || lx == rx)
		    break;

		  if (array)
		    replace_nokill(sset, sset, ssln, lx, rx);
		  else
		    replace_kill(sset, sset, ssln, lx, rx); 

		  break;
		  
		case FNAME:
		  
		  if (array)
		    alias_nokill(sset, sset, ssln, lx, rx);
		  else
		    alias_kill(sset, sset, ssln, lx, rx);
		}
	    }    /* assignment to function pointer */

	break; /* ASGN */
      }

    case NEW:
      
      array = el->n->lft->lft ? 1 : 0;
      lx    = el->n->lft->idx;
      
      if (array)
	alias_nokill(sset, sset, ssln, lx, HEAP);
      else
	alias_kill(sset, sset, ssln, lx, HEAP);

      break; /* NEW */
      
    case FUNCALL:
      {	
	int ntyp = el->n->rgt->ntyp;

	Lextok *arg = el->n->rgt->lft;
	Lextok *ret = el->n->lft;  

	fx = el->n->rgt->rgt->idx;

	switch (ntyp)
	  {
	  case NAME:
	    {
	      int i;

	      for (i = F_MIN; i < F_MAX; i ++)
		{
		  if (set_mem(sset->s_set [fx], i))
		    {
		      call_merge(arg, i, sset, ssln);
		      retn_merge(ret, i, sset, ssln);
		    }
		}

	      break;
	    }
	    
	  case FNAME:

	    call_merge(arg, fx, sset, ssln);
	    retn_merge(ret, fx, sset, ssln);
	  }
	
	break; /* FUNCALL */
      }

    case RETURN:
      {	
	int idx;

	fx = el->n->idx;	

	if (el->n->lft)
	  {
	    int ntyp = el->n->lft->ntyp;

	    switch (ntyp)
	      {
	      case LVAL:

		idx = el->n->lft->lft->idx;		
		set_add(F_TAB(fx)->rset, idx);
	      
		break;

	      case NAME:

		idx = el->n->lft->idx;
		
		if (V_TAB(idx)->ptr
		    || V_TAB(idx)->fptr)
		  set_add(F_TAB(fx)->rset, idx);

		break;

	      case FNAME:

		idx = el->n->lft->idx;
		set_add(F_TAB(fx)->rset, idx);
	      }
	  }
	
	ss_merge(F_TAB(fx)->out, sset);
	break; /* RETURN */
      }

    case RETFUN:
      {
	int idx  = el->n->idx;
	int ntyp = el->n->rgt->ntyp;
	Lextok *arg = el->n->rgt->lft;

	fx = el->n->rgt->rgt->idx;

	switch (ntyp)
	  {
	  case NAME:
	    {
	      int i;

	      for (i = F_MIN; i < F_MAX; i ++)
		{
		  if (set_mem(sset->s_set [fx], i))
		    {
		      call_merge(arg, i, sset, ssln);		   
		      set_union(F_TAB(idx)->rset, F_TAB(i)->rset);
		      ss_merge(F_TAB(idx)->out, F_TAB(i)->out);
		    }
		}

	      break;
	    }

	  case FNAME:

	    call_merge(arg, fx, sset, ssln);
	    set_union(F_TAB(idx)->rset, F_TAB(fx)->rset);	      
	    ss_merge(F_TAB(idx)->out, F_TAB(fx)->out);
	  }
	
	break; /* RETFUN */
      }
    }
}


static void trans_bb(bb_t block)
{
  int i;

  ss_copy(block->out, block->in);
  ss_copy(block->out_ln, block->in_ln);

  for (i = 0; i < block->e_no; i ++)
    trans_el(block->e_seq [i], block->out, block->out_ln);
}


static void merge_bb(bb_t block)
{
  int i;

  for (i = 0; i < block->pred_no; i ++)
    {
      ss_merge(block->in, block->pred_list [i]->out);
      ss_merge(block->in_ln, block->pred_list [i]->out_ln);
    }
}


static void initsets(bl_t bl)
{
  int i;

  for (i = 0; i < NO(bl); i ++)
    {
      bb_t block = LS(bl) [i];

      block->in  = a_create();
      block->out = a_create();

      block->in_ln  = l_create();
      block->out_ln = l_create();
    }
}


static int iterate(bl_t bl)
{
  bl_t wlist  = bl_create();
  ss_t oldout = a_create();
  
  int result = 0, i;

  for (i = 0; i < NO(bl); i ++)
    {
      INC(LS(wlist), bb_t, NO(wlist));
      LS(wlist) [i] = LS(bl) [i];
    }

  while (NO(wlist))
    {
      bb_t node = LS(wlist) [NO(wlist) - 1];

      DEC(LS(wlist), bb_t, NO(wlist));

#if DEBUG>=1
      printf(".");
      fflush(stdout);
#endif 

      merge_bb(node);
      ss_copy(oldout, node->out);
      trans_bb(node);

      if (!ss_equal(node->out, oldout))
	{
	  result ++;

	  for (i = 0; i < node->succ_no; i ++)
	    {
	      bb_t s = node->succ_list [i];
	      int j;

	      for (j = 0; j < NO(wlist) && LS(wlist) [j]->bid != s->bid; j ++);
	      
	      if (j == NO(wlist))
		{
		  INC(LS(wlist), bb_t, NO(wlist));
		  LS(wlist) [NO(wlist) - 1] = s;
		}
	    }
	}
    }

  ss_delete(oldout);
  bl_delete(wlist);

  return result;
}


static void derive_sets(Element *e, ss_t sset, ss_t ssln)
{
  bb_t block = e->block;
  int i;

  ss_copy(sset, block->in);
  ss_copy(ssln, block->in_ln);

  for (i = 0; i < block->e_no; i ++)
    {
      if (block->e_seq [i] == e)
	break;
      
      trans_el(block->e_seq [i], sset, ssln);
    }
}


/*** Interprocedural dataflow routines ***/

static void defun_part(Defun *fn)
{
  Fdx = fn->idx;

  bb_leader(fn->bb_list, fn->body->frst, 1);
  bb_setup(NULL, fn->body->frst);

  bb_link(fn->bb_list);

  bb_order(LS(fn->bb_list) [0], fn->bb_dfn);
  bb_reach(fn->bb_list, fn->fun->name);
}


static void proc_part(ProcList *p)
{
  enter_seq(p->s);

  bb_leader(p->bb_list, p->s->frst, 1);
  bb_setup(NULL, p->s->frst);

  bb_link(p->bb_list);

  bb_order(LS(p->bb_list) [0], p->bb_dfn);
  bb_reach(p->bb_list, p->n->name);

  exit_seq();
}


/*** Exported routines ***/

int sym_idx()
{
  return s_idx ++;
}


int fun_idx(Defun *fn)
{
  int i;

  for (i = 0; i < f_idx; i ++)
    {
      if (f_tab [i]->fun == fn)
	return i;
    }

  INC(f_tab, fun_t, f_idx);

  f_tab [i] = (fun_t) emalloc (sizeof(struct fun_tag));  
  f_tab [i]->fun    = fn;

  fn->idx = i;

  return i;
}


void dfa_all(ProcList *proclist, Defun *funlist)
{
  ProcList *p;
  Defun *fn;
  int change = 1;

  if (no_dfa)
    return;

  for (fn = funlist; fn; fn = fn->nxt)
    {
      fn->bb_list = bl_create();
      fn->bb_dfn  = bl_create();
  
      enter_seq(fn->body);
      
      formal_idx(fn->prms);
      init_idx(fn->body);

      defun_part(fn);
      
      exit_seq();
    }

  for (p = proclist; p; p = p->nxt)
    {
      p->bb_list = bl_create();
      p->bb_dfn  = bl_create();

      init_idx(p->s);

      proc_part(p);
    }

  for (fn = funlist; fn; fn = fn->nxt)
    {
      bb_t frst = LS(fn->bb_dfn) [NO(fn->bb_dfn) - 1];
      int fx = fn->idx;

      initsets(fn->bb_list);
      
      F_TAB(fx)->in    = frst->in;
      F_TAB(fx)->in_ln = frst->in_ln;

      F_TAB(fx)->out   = a_create();
      F_TAB(fx)->rset  = set_create(V_MAX);

      scope_merge(fn->body, frst->in, frst->in_ln);
    }

  for (p = proclist; p; p = p->nxt)
    {
      bb_t frst = LS(p->bb_dfn) [NO(p->bb_dfn) - 1];

      initsets(p->bb_list);

      scope_merge(p->s, frst->in, frst->in_ln);
    }

  while (change)
    {
      change = 0;

      for (p = proclist; p; p = p->nxt) 
	{
#if DEBUG>=1
	  printf("iterate [%s]", p->n->name);
	  fflush(stdout);
#endif
	  iterate(p->bb_dfn);
#if DEBUG>=1
	  printf("\n");
#endif
	}

      for (fn = funlist; fn; fn = fn->nxt)
	{
#if DEBUG>=1
	  printf("iterate [%s]", fn->fun->name);
	  fflush(stdout);
#endif
	  change += iterate(fn->bb_dfn);
#if DEBUG>=1
	  printf("\n");
#endif
	}
    }

#if DEBUG>=2
  for (fn = funlist; fn; fn = fn->nxt)
    {
      printf("function %s()\n\n", fn->fun->name);
      bb_printf(fn->bb_dfn);
    }

  for (p = proclist; p; p = p->nxt)
    {
      printf("proctype %s()\n\n", p->n->name);
      bb_printf(p->bb_dfn);
    }
#endif /* DEBUG */
}


int global_alias(Element *e, Lextok *n)
{
  ss_t sset, ssln;
  int i;

  if (strcmp(n->sym->name, "_pid") == 0
      || strcmp(n->sym->name, "_p") == 0
      || strcmp(n->sym->name, "_") == 0)
    return 0;

  if (no_dfa)
    {
      Symbol *s = findloc2(n->sym);

      if (!s)
	s = n->sym;

      if (s->type & POINTER)
	return 1;

      return s->context ? 0 : 1;
    }

  if (V_TAB(n->idx)->glob)
    return 1;

  sset = a_create();
  ssln = l_create();

  derive_sets(e, sset, ssln);

  for (i = V_MIN; i < V_MAX; i ++)
    {
      if (i == n->idx)
	continue;

      if (set_mem(sset->s_set [n->idx], i)
	  && V_TAB(i)->glob)
	{
	  ss_delete(sset);
	  ss_delete(ssln);

	  return 1;
	}
    }

  if (set_mem(sset->s_set [n->idx], HEAP))
    {
      ss_delete(sset);
      ss_delete(ssln);

      return 1;      
    }

  for (i = V_MIN; i < V_MAX; i ++)
    {
      if (i == n->idx)
	continue;

      if (set_mem(ssln->s_set [n->idx], i)
	  && V_TAB(i)->glob)
	{
	  ss_delete(sset);
	  ss_delete(ssln);

	  return 1;
	}
    }

  return 0;
}
