/***** spin: flow.c *****/

/* Copyright (c) 1991-1998 by Lucent Technologies - Bell Laboratories     */
/* All Rights Reserved.  This software is for educational purposes only.  */
/* Permission is given to distribute this code provided that this intro-  */
/* ductory message is not removed and no monies are exchanged.            */
/* No guarantee is expressed or implied by the distribution of this code. */
/* Software written by Gerard J. Holzmann as part of the book:            */
/* `Design and Validation of Computer Protocols,' ISBN 0-13-539925-4,     */
/* Prentice Hall, Englewood Cliffs, NJ, 07632.                            */
/* Send bug-reports and/or questions to: gerard@research.bell-labs.com    */

/* dSpin extension written by Radu Iosif. Feedback: iosif@athena.polito.it */

#include <stdlib.h>

#include "spin.h"
#ifdef PC
#include "y_tab.h"
#else
#include "y.tab.h"
#endif

extern Symbol	*Fname, *context;
extern int	nr_errs, lineno, verbose;
extern short	has_unless, has_badelse;

Element *Al_El = ZE;
Label	*labtab = (Label *) 0;
int	 Elcnt = 0, DstepStart = -1;
SeqList	*cur_s = (SeqList *) 0;

static int  ScopeNo = 0;
static int *ScopeStart = NULL;

static Lbreak	*breakstack = (Lbreak *) 0;
static Lextok	*innermost;
static int	Unique=0, break_id=0;
static int      ScopeDepth = 0;

static Element	*if_seq(Lextok *);
static Element	*new_el(Lextok *);
static Element	*unless_seq(Lextok *);
static void	add_el(Element *, Sequence *);
static void	attach_escape(Sequence *, Sequence *);
static void	mov_lab(Symbol *, Element *, Element *);
static void	set_lab(Symbol *, Element *);
static void	walk_atomic(Element *, Element *, int);

extern Ordered *all_names;

Symbol	       *has_lab(Element *);


void
insert_symbol(Sequence *sq, Symbol  *s)
{	
  if (!sq->symtab)
    {
      s->prev = s->next = ZS;
      sq->symtab = sq->stail = s;
      return;
    }

  sq->stail->next = s;
  s->prev = sq->stail;
  sq->stail = s;
}


void
setlocals(Sequence *sq)
{	
  Ordered *walk, *next;
  Symbol  *sp;

  for (walk = all_names; walk; walk = next)
    {	
      sp = walk->entry;
      next = walk->next;

      if (sp
	  && !sp->owner
	  && (sp->context == context)
	  && !(sp->hidden & 128)
	  && (sp->Nid >= 0)
	  && (sp->type == UNSIGNED
	      || sp->type == BIT
	      || sp->type == MTYPE
	      || sp->type == BYTE
	      || sp->type == CHAN
	      || sp->type == SHORT
	      || sp->type == INT
	      || sp->type == STRUCT
	      || sp->type == FTYPE
	      || sp->type == LABEL 
	      || sp->type & POINTER))
	{
	  delsym(sp);               /* delete the entry */
	  insert_symbol(sq, sp);    /* insert it into sequence */ 
	}
    }
}


static char *
new_scope_name(char *o)
{
  static int x = 0;
  char * s = (char *) emalloc(strlen(o) + 32);

  sprintf(s, "%s$%d", o, x ++);

  return s;
}


Element *
mk_skip(void)
{	
  Lextok  *t = nn(ZN, CONST, ZN, ZN);

  t->val = 1;
  return new_el(nn(ZN, 'c', t, ZN));
}


static int
may_block(Lextok *now)
{
  switch (now->ntyp)
    {
    case 'c':
      return may_block(now->lft);

    case CONST:
      return now->val == 0;

    case TIMEOUT: case RUN:    case NEW:    case DELETE:
    case FUNCALL: case RETFUN: case RETURN: case ELSE:
    case ASGN:    case PRINT:  case ASSERT: case '.':
    case BREAK:   case GOTO:   case STOP:
      return 0;

    default:
      return 1;
    }
}


void
rem_local(void)
{
  ScopeNo ++;
  ScopeStart = (int *) erealloc(ScopeStart, ScopeNo * sizeof(int));
  ScopeStart [ScopeNo - 1] = Unique;
}


void unrem_local(void)
{
  ScopeNo --;
  ScopeStart = (int *) erealloc(ScopeStart, ScopeNo * sizeof(int));
}


static void
mark_jumps(Sequence *s, Element *start, Element *stop)
{
  Element *l, *e;
  SeqList *sub;
  
  for (e = start; e; e = e->nxt)
    {
      if (e && e->n)
	{
	  switch (e->n->ntyp)
	    {
	    case NON_ATOMIC:
	      mark_jumps(s, e->n->sl->this->frst, e->n->sl->this->last);
	      break;

	    case GOTO:
	      l = get_lab(e->n, 0);
	      if (!l || l->Seqno < ScopeStart[ScopeNo - 1])
		{
		  e->tag |= JMP;
		  e->sel = seqlist(s, e->sel);
		}

	      break;

	    case IF:
	    case DO:
	    case UNLESS:
	      for (sub = e->sub; sub; sub = sub->nxt)
		mark_jumps(s, sub->this->frst, sub->this->last);
	    }
	}
	 
      if (e == stop)
	break;
    }
  
}


void
open_seq(char *name, int top, int local, int block)
{	
  Sequence *s = (Sequence *) emalloc(sizeof(Sequence));
  char *old = "";

  cur_s = seqlist(s, cur_s);
  s->block = block;

  if (top) 
    Elcnt = 1;

  if (context)
    old = context->name;

  if (local) 
    {
      s->cnxt = context;
      context = (Symbol *) emalloc(sizeof(Symbol));

      if (name)
	context->name = name;
      else
	context->name = new_scope_name(old);

      ScopeDepth ++;
    }
}


Sequence *
close_seq(int nottop, int local)
{	
  Sequence *s = cur_s->this;

  if (!s->last)
    fatal("empty sequence", 0);

  if (local) 
    {
      setlocals(s);
      context = cur_s->this->cnxt;
      ScopeDepth --;
    }
  
  cur_s = cur_s->nxt;

  if (s->last->n->ntyp == FUNCALL)
    add_el(mk_skip(), s);

  if (s->block &&
      (s->last->n->ntyp == BREAK || 
       s->last->n->ntyp == '.'   ||
       may_block(s->last->n)))
    add_el(mk_skip(), s);

  s->maxel = Elcnt;
  s->extent = s->last;

  if (s->block)
    mark_jumps(s, s->frst, s->last);

  return s;
}


void
enter_seq(Sequence *s)
{
  cur_s = seqlist(s, cur_s);
}


void
exit_seq()
{
  cur_s = cur_s->nxt;
}


Element *
frst_el(Sequence *s)
{
  switch (s->frst->n->ntyp)
    {
    case D_STEP:
    case ATOMIC:
    case NON_ATOMIC:
      return frst_el(s->frst->n->sl->this);

    default:
      return s->frst;
    }
}


Element *
last_el(Sequence *s)
{
  switch (s->last->n->ntyp)
    {
    case D_STEP:
    case ATOMIC:
    case NON_ATOMIC:
      return last_el(s->last->n->sl->this);

    default:
      return s->last;
    }
}


void
rem_Seq(void)
{
  DstepStart = Unique;
}


void
unrem_Seq(void)
{
  DstepStart = -1;
}


static int
Rjumpslocal(Element *start, Element *stop)
{	
  Element *lb, *f;
  SeqList *h;

  /* allow no jumps out of a d_step sequence */
  for (f = start; f && f != stop; f = f->nxt)
    {	
      if (f && f->n && f->n->ntyp == GOTO)
	{	
	  lb = get_lab(f->n, 0);

	  if (!lb || lb->Seqno < DstepStart)
	    {	
	      lineno = f->n->ln;
	      Fname = f->n->fn;
	      return 0;
	    }	
	}
      
      for (h = f->sub; h; h = h->nxt)
	{	
	  if (!Rjumpslocal(h->this->frst, h->this->last))
	    return 0;	  
	}
    }
	
  return 1;
}


void
cross_dsteps(Lextok *a, Lextok *b)
{
	if (a && b
	&&  a->indstep != b->indstep)
	{	lineno = a->ln;
		Fname  = a->fn;
		fatal("jump into d_step sequence", (char *) 0);
	}
}

Sequence *
check_seq(Sequence *s, int nottop)
{
  Symbol *z;

  if (nottop > 0 && (z = has_lab(s->frst)))
    {	
      printf("error: (%s:%d) label %s placed incorrectly\n",
	     (s->frst->n)?s->frst->n->fn->name:"-",
	     (s->frst->n)?s->frst->n->ln:0,
	     z->name);

      switch (nottop) {
      case 1:
	printf("=====> stmnt unless Label: stmnt\n");
	printf("sorry, cannot jump to the guard of an\n");
	printf("escape (it is not a unique state)\n");
	break;

      case 2:
	printf("=====> instead of  ");
	printf("\"Label: stmnt unless stmnt\"\n");
	printf("=====> always use  ");
	printf("\"Label: { stmnt unless stmnt }\"\n");
	break;

      case 3:
	printf("=====> instead of  ");
	printf("\"atomic { Label: statement ... }\"\n");
	printf("=====> always use  ");
	printf("\"Label: atomic { statement ... }\"\n");
	break;

      case 4:
	printf("=====> instead of  ");
	printf("\"d_step { Label: statement ... }\"\n");
	printf("=====> always use  ");
	printf("\"Label: d_step { statement ... }\"\n");
	break;

      case 5:
	printf("=====> instead of  ");
	printf("\"{ Label: statement ... }\"\n");
	printf("=====> always use  ");
	printf("\"Label: { statement ... }\"\n");
	break;

      case 6:
	printf("=====>instead of\n");
	printf("	do (or if)\n");
	printf("	:: ...\n");
	printf("	:: Label: statement\n");
	printf("	od (of fi)\n");
	printf("=====>always use\n");
	printf("Label:	do (or if)\n");
	printf("	:: ...\n");
	printf("	:: statement\n");
	printf("	od (or fi)\n");
	break;
      }

      alldone(1);
    }

  if (nottop == 4
      && !Rjumpslocal(s->frst, s->last))
    fatal("non_local jump in d_step sequence", (char *) 0);

  if (nottop == 7)
    {           
      if (s->frst->n->ntyp == NON_ATOMIC)
	{
	  printf("=====>instead of\n");
	  printf("        atomic (or d_step) {\n");
	  printf("             { scope } ... }\n");
	  printf("=====>always use\n");
	  printf("        { atomic (or d_step)\n");
	  printf("             { scope } ... }\n");
	  alldone(1);
	}
    }

  return s;
}


Lextok *
do_unless(Lextok *No, Lextok *Es)
{	
  SeqList *Sl;
  Lextok *Re = nn(ZN, UNLESS, ZN, ZN);
  Re->ln = No->ln;
  Re->fn = No->fn;

  has_unless++;

  if (Es->ntyp == NON_ATOMIC)
    Sl = Es->sl;
  else
    {	
      open_seq(0, 0, 0, 0); add_seq(Es);
      Sl = seqlist(close_seq(0, 0), 0);
    }

  if (No->ntyp == NON_ATOMIC)
    {	
      No->sl->nxt = Sl;
      Sl = No->sl;
    } else	
      if (No->ntyp == ':'
	  && (No->lft->ntyp == NON_ATOMIC
	      ||  No->lft->ntyp == ATOMIC
	      ||  No->lft->ntyp == D_STEP))
	{
	  int tok = No->lft->ntyp;
	  
	  No->lft->sl->nxt = Sl;
	  Re->sl = No->lft->sl;
	  
	  open_seq(0, 0, 1, 1); 
	  add_seq(Re);
	  Re = nn(ZN, tok, ZN, ZN);
	  Re->sl = seqlist(close_seq(7, 1), 0);
	  Re->ln = No->ln;
	  Re->fn = No->fn;
	  
	  Re = nn(No, ':', Re, ZN);	/* lift label */
	  Re->ln = No->ln;
	  Re->fn = No->fn;
	  return Re;
	} 
      else
	{	
	  open_seq(0, 0, 1, 1); 
	  add_seq(No);
	  Sl = seqlist(close_seq(2, 1), Sl);
	}
  
  Re->sl = Sl;
  return Re;
}


SeqList *
seqlist(Sequence *s, SeqList *r)
{	
  SeqList *t = (SeqList *) emalloc(sizeof(SeqList));
  
  t->this = s;
  t->nxt = r;
  
  return t;
}


static Element *
new_el(Lextok *n)
{	
  Element *m;

  if (n)
    {	
      if (n->ntyp == IF || n->ntyp == DO)
	return if_seq(n);
	
      if (n->ntyp == UNLESS)
	return unless_seq(n);
    }
	
  m = (Element *) emalloc(sizeof(Element));
  m->n = n;
  m->seqno = Elcnt++;
  m->Seqno = Unique++;
  m->Nxt = Al_El; Al_El = m;
  return m;
}


static int
has_chanref(Lextok *n)
{
	if (!n) return 0;

	switch (n->ntyp) {
	case 's':	case 'r':
#if 0
	case 'R':	case LEN:
#endif
	case FULL:	case NFULL:
	case EMPTY:	case NEMPTY:
		return 1;
	default:
		break;
	}
	if (has_chanref(n->lft))
		return 1;

	return has_chanref(n->rgt);
}


static Element *
if_seq(Lextok *n)
{	
  int	tok = n->ntyp;
  SeqList	*s  = n->sl;
  Element	*e  = new_el(ZN);
  Element	*t  = new_el(nn(ZN,'.',ZN,ZN)); /* target */
  SeqList	*z, *prev_z = (SeqList *) 0;
  SeqList *move_else  = (SeqList *) 0;	/* to end of optionlist */
  int	ref_chans = 0;

  for (z = s; z; z = z->nxt)
    {	
      if (!z->this->frst)
	continue;
	
      if (z->this->frst->n->ntyp == ELSE)
	{	
	  if (move_else)
	    fatal("duplicate `else'", (char *) 0);
		
	  if (z->nxt)	/* is not already at the end */
	    {	
	      move_else = z;
	      if (prev_z)
		prev_z->nxt = z->nxt;
	      else
		s = n->sl = z->nxt;
	      continue;
	    }
	} else
	  ref_chans |= has_chanref(z->this->frst->n);
      prev_z = z;
    }
	
  if (move_else)
    {	
      move_else->nxt = (SeqList *) 0;
      /* if there is no prev, then else was at the end */
      if (!prev_z) fatal("cannot happen - if_seq", (char *) 0);
      prev_z->nxt = move_else;
      prev_z = move_else;
    }
	
  if (prev_z
      &&  ref_chans
      &&  prev_z->this->frst->n->ntyp == ELSE)
    {	
      prev_z->this->frst->n->val = 1;
      has_badelse++;
      non_fatal("dubious use of 'else' combined with i/o,", (char *)0);
      nr_errs--;
    }

  e->n = nn(n, tok, ZN, ZN);
  e->n->sl = s;			/* preserve as info only */
  e->sub = s;

  for (z = s; z; prev_z = z, z = z->nxt)
    add_el(t, z->this);	/* append target */
	
  if (tok == DO)
    {	
      add_el(t, cur_s->this); /* target upfront */
      t = new_el(nn(n, BREAK, ZN, ZN)); /* break target */
      set_lab(break_dest(), t);	/* new exit  */
      breakstack = breakstack->nxt;	/* pop stack */
    }
	
  add_el(e, cur_s->this);
  add_el(t, cur_s->this);

  return e;			/* destination node for label */
}


static void
escape_el(Element *f, Sequence *e)
{	SeqList *z;

 f->esc = seqlist(e, f->esc);	/* lifo order... */

#if (DEBUG >= 1)
 printf("attach %d (", e->frst->Seqno);
 comment(stdout, e->frst->n, 0);
 printf(")	to %d (", f->Seqno);
 comment(stdout, f->n, 0);
 printf(")\n");
#endif

 switch (f->n->ntyp) 
   {
   case UNLESS:
     attach_escape(f->sub->this, e);
     break;
   
   case IF:
   case DO:
     for (z = f->sub; z; z = z->nxt)
       attach_escape(z->this, e);
     break;
	
   case D_STEP:
     /* attach only to the guard stmnt */
     escape_el(f->n->sl->this->frst, e);
     break;
	
   case ATOMIC:
   case NON_ATOMIC:
     /* attach only to all stmnts */
     attach_escape(f->n->sl->this, e);
     break;
   }
}


static void
attach_escape(Sequence *n, Sequence *e)
{	
  Element *f;
  
  for (f = n->frst; f; f = f->nxt)
    {	
      escape_el(f, e);
      if (f == n->extent)
	break;
    }
}


static Element *
unless_seq(Lextok *n)
{	SeqList	*s  = n->sl;
	Element	*e  = new_el(ZN);
	Element	*t  = new_el(nn(ZN,'.',ZN,ZN)); /* target */
	SeqList	*z;

	e->n = nn(n, UNLESS, ZN, ZN);
	e->n->sl = s;			/* info only */
	e->sub = s;

	/* need 2 sequences: normal execution and escape */
	if (!s || !s->nxt || s->nxt->nxt)
		fatal("unexpected unless structure", (char *)0);

	/* append the target state to both */
	for (z = s; z; z = z->nxt)
		add_el(t, z->this);

	/* attach escapes to all states in normal sequence */
	attach_escape(s->this, s->nxt->this);

	add_el(e, cur_s->this);
	add_el(t, cur_s->this);
#if (DEBUG >= 1)
	printf("unless element (%d,%d):\n", e->Seqno, t->Seqno);
	for (z = s; z; z = z->nxt)
	{	Element *x; printf("\t%d,%d,%d :: ",
		z->this->frst->Seqno,
		z->this->extent->Seqno,
		z->this->last->Seqno);
		for (x = z->this->frst; x; x = x->nxt)
			printf("(%d)", x->Seqno);
		printf("\n");
	}
#endif
	return e;
}


static void
add_el(Element *e, Sequence *s)
{
  if (e->n->ntyp == GOTO)
    {	
      Symbol *z = has_lab(e);
	
      if (z
	  && (strncmp(z->name, "progress", 8) == 0
	      ||  strncmp(z->name, "accept", 6) == 0
	      ||  strncmp(z->name, "end", 3) == 0))
	{	
	  Element *y; /* insert a skip */

	  y = mk_skip();
	  mov_lab(z, e, y); /* inherit label */
	  add_el(y, s);
	}	
    }  

#if (DEBUG >= 1)
  printf("add_el %d after %d -- ", e->Seqno, (s->last)?s->last->Seqno:-1);
  comment(stdout, e->n, 0);
  printf("\n");
#endif

  if (s->block && 
      !s->frst   &&
      (e->n->ntyp == '.' ||
       e->n->ntyp == IF  ||
       may_block(e->n)))
    add_el(mk_skip(), s);
  
  if (!s->frst)
    s->frst = e;
  else
    s->last->nxt = e;
  
  s->last = e;
}


static Element *
colons(Lextok *n)
{
  if (!n)
    return ZE;

  if (n->ntyp == ':')
    {	
      Element *e = colons(n->lft);
      set_lab(n->sym, e);
      return e;
    }

  innermost = n;
  return new_el(n);
}


void
add_seq(Lextok *n)
{	
  Element *e;
  SeqList *l;

  if (!n) 
    return;
  
  innermost = n;
  e = colons(n);

  if (innermost->ntyp != IF
      && innermost->ntyp != DO
      && innermost->ntyp != UNLESS)
    add_el(e, cur_s->this);
}


static void
set_lab(Symbol *s, Element *e)
{	
  extern Symbol *context;
  Label *l; 

  if (!s) 
    return;

  for (l = labtab; l; l = l->nxt)
    {
      if (l->c == context &&
	  strcmp(l->s->name, s->name) == 0)
	{	
	  non_fatal("label '%s' redeclared", s->name);
	  break;
	}
    }

  if (verbose&32 && ScopeDepth>1) 
    printf("spin: line %3d %s, Warning: Label '%s' in inner local scope\n", 
	   lineno, Fname ? Fname->name : "", s->name);

  l = (Label *) emalloc(sizeof(Label));

  l->s = s;
  l->c = context;
  l->e = e;
  l->nxt = labtab;
  labtab = l;
}


Element *
get_lab(Lextok *n, int md)
{	
  Label *l;
  Symbol *s = n->sym;

  /* I really hope that just a simple name check will do ! */
  for (l = labtab; l; l = l->nxt)
    {
      if (strcmp(s->name, l->s->name) == 0)
	return l->e;
    }

  lineno = n->ln;
  Fname = n->fn;
  
  if (md) 
    fatal("undefined label %s", s->name);
  
  return ZE;
}


Symbol *
has_lab(Element *e)
{	
  Label *l;

  for (l = labtab; l; l = l->nxt)
    if (e == l->e)
      return (l->s);
	
  return ZS;
}

static void
mov_lab(Symbol *z, Element *e, Element *y)
{	
  Label *l;

  for (l = labtab; l; l = l->nxt)
    if (e == l->e)
      {	
	l->e = y;
	return;
      }

  if (e->n)
    {	
      lineno = e->n->ln;
      Fname  = e->n->fn;
    }
  
  fatal("cannot happen - mov_lab %s", z->name);
}

void
fix_dest(Symbol *c, Symbol *a)	/* label, proctype */
{	
  Label *l;
	
  for (l = labtab; l; l = l->nxt)    
    if (strcmp(c->name, l->s->name) == 0 
	&& strcmp(a->name, l->c->name) == 0)
      break;
	
  if (!l)
    {	
      non_fatal("unknown label '%s'", c->name);
      return;
    }
	
  if (!l->e || !l->e->n)
    fatal("fix_dest error (%s)", c->name);
	
  if (l->e->n->ntyp == GOTO)
    {	
      Element	*y = (Element *) emalloc(sizeof(Element));
      int	keep_ln = l->e->n->ln;
      Symbol	*keep_fn = l->e->n->fn;
      
      /* insert skip - or target is optimized away */
      y->n = l->e->n;		  /* copy of the goto   */
      y->seqno = find_maxel(a); /* unique seqno within proc */
      y->nxt = l->e->nxt;
      y->Seqno = Unique++; y->Nxt = Al_El; Al_El = y;
     
      /* turn the original element+seqno into a skip */
      l->e->n = nn(ZN, 'c', nn(ZN, CONST, ZN, ZN), ZN);
      l->e->n->ln = l->e->n->lft->ln = keep_ln;
      l->e->n->fn = l->e->n->lft->fn = keep_fn;
      l->e->n->lft->val = 1;
      l->e->nxt = y;		/* append the goto  */
    }
  
  l->e->status |= CHECK2;	/* treat as if global */
}

int
find_lab(Symbol *s, Symbol *c, int markit)
{	Label *l;

	for (l = labtab; l; l = l->nxt)
	{	if (strcmp(s->name, l->s->name) == 0
		&&  strcmp(c->name, l->c->name) == 0)
		{	l->visible |= markit;
			return (l->e->seqno);
	}	}
	return 0;
}

void
pushbreak(void)
{	Lbreak *r = (Lbreak *) emalloc(sizeof(Lbreak));
	Symbol *l;
	char buf[32];

	sprintf(buf, ":b%d", break_id++);
	l = lookup(buf);
	r->l = l;
	r->nxt = breakstack;
	breakstack = r;
}

Symbol *
break_dest(void)
{
	if (!breakstack)
		fatal("misplaced break statement", (char *)0);
	return breakstack->l;
}

void
make_atomic(Sequence *s, int added)
{
	walk_atomic(s->frst, s->last, added);
	s->last->status &= ~ATOM;
	s->last->status |= L_ATOM;
}

static void
walk_atomic(Element *a, Element *b, int added)
{	Element *f; Symbol *ofn; int oln;
	SeqList *h;

	ofn = Fname;
	oln = lineno;
	for (f = a; ; f = f->nxt)
	{	f->status |= (ATOM|added);
		switch (f->n->ntyp) {
		case ATOMIC:
			printf("warning: line %3d %s, atomic inside %s (ignored)\n",
			  f->n->ln, f->n->fn->name, (added)?"d_step":"atomic");
			f->n->ntyp = NON_ATOMIC;
			goto mknonat;
		case D_STEP:
			printf("warning: line %3d %s, d_step inside ",
			 f->n->ln, f->n->fn->name);
			if (added)
			{	printf("d_step (ignored)\n");
				f->n->ntyp = NON_ATOMIC;
				goto mknonat;
			}
			printf("atomic\n");
			break;
		case NON_ATOMIC:
mknonat:		h = f->n->sl;
			walk_atomic(h->this->frst, h->this->last, added);
			break;
		}
		for (h = f->sub; h; h = h->nxt)
			walk_atomic(h->this->frst, h->this->last, added);
		if (f == b)
			break;
	}
	Fname = ofn;
	lineno = oln;
}

void
dumplabels(void)
{	Label *l;

	for (l = labtab; l; l = l->nxt)
		if (l->c != 0 && l->s->name[0] != ':')
		printf("label	%s	%d	<%s>\n",
		l->s->name, l->e->seqno, l->c->name);
}
